"""Callbacks that create, display and store animation according to dynamic data."""

import subprocess
import time
from abc import ABC, abstractmethod
from collections.abc import Iterable
from contextlib import contextmanager
from copy import copy
from multiprocessing import Process
from unittest.mock import Mock
import matplotlib
from matplotlib.patches import Rectangle
from matplotlib.ticker import MaxNLocator

matplotlib.use("Agg")
import matplotlib.animation
import matplotx.styles
from matplotlib.axes import Axes
from matplotlib.gridspec import GridSpec, GridSpecFromSubplotSpec, SubplotSpec
from matplotlib.transforms import Affine2D

import regelum
import os


import matplotlib.pyplot as plt


from pathlib import Path

import regelum.__internal.base

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure
import numpy as np
from matplotlib.backends import backend_qt5agg  # e.g.

from svgpathtools import svg2paths
from svgpath2mpl import parse_path

import regelum.callback as callback


plt.rcParams["animation.frame_format"] = "svg"  # VERY important

import matplotlib.style as mplstyle

plt.style.use(matplotx.styles.dracula)
# plt.style.use('fast')
# plt.rcParams['path.simplify'] = True

# plt.rcParams['path.simplify_threshold'] = 1.0


class AnimationCallback(callback.Callback, ABC):
    """Callback (base) responsible for rendering animated visualizations of the experiment."""

    _frames = 100
    _is_global = False

    @classmethod
    def _compose(cls_left, cls_right):
        animations = []
        for cls in [cls_left, cls_right]:
            if cls is None:
                continue
            if issubclass(cls, ComposedAnimationCallback) and not issubclass(
                cls, DeferredComposedAnimation
            ):
                assert (
                    cls is not ComposedAnimationCallback
                ), "Adding a composed animation in this way is ambiguous."
                animations += cls._animations
            elif issubclass(cls, AnimationCallback):
                animations.append(cls)
            else:
                raise TypeError(
                    "An animation class can only be added with another animation class or None."
                )

        class Composed(ComposedAnimationCallback):
            _animations = animations

            def __init__(self, **kwargs):
                super().__init__(*self._animations, **kwargs)

        return Composed

    def __init__(self, *args, interactive=None, fig=None, ax=None, mng=None, **kwargs):
        """Initialize an instance of AnimationCallback."""
        super().__init__(*args, **kwargs)
        # assert (
        #    self.attachee is not None
        # ), "An animation callback can only be instantiated via attachment, however an attempt to instantiate it otherwise was made."
        self.frame_data = []
        # matplotlib.use('Qt5Agg', force=True)

        self.interactive_mode = (
            self._metadata["argv"].interactive if interactive is None else interactive
        )
        if self.interactive_mode:
            matplotlib.use("Agg")
            self.fig = Figure(figsize=(10, 10))
            canvas = FigureCanvas(self.fig)

            self.ax = canvas.figure.add_subplot(111)
            self.mng = backend_qt5agg.new_figure_manager_given_figure(1, self.fig)
            self.setup()
        else:
            self.mng = mng

            self.fig, self.ax = fig, ax
            if isinstance(self.ax, SubplotSpec):
                self.ax = fig.add_subplot(self.ax)
        self.save_directory = Path(
            f".callbacks/{self.__class__.__name__}"
            + (f"@{self.attachee.__name__}" if self.attachee is not None else "")
        ).resolve()
        self.saved_counter = 0
        self.skip_frames = self._metadata["skip_frames"]
        self.counter = 0

    def __del__(self):
        if hasattr(self, "updater"):
            self.canvas_shm["canvas"] = None
            self.updater.join()

    @abstractmethod
    def setup(self):
        pass

    def get_save_directory(self):
        return str(self.save_directory)

    def add_frame(self, **frame_datum):
        if self.counter is not None:
            self.counter += 1
            if (self.counter - 1) % self.skip_frames:
                return
        self.frame_data.append(frame_datum)
        if hasattr(self, "interactive_status"):
            self.interactive_status["frame"] = frame_datum

        if self.interactive_mode:
            self.lim(frame_idx=len(self.frame_data) - 1)
            self.construct_frame(**frame_datum)
            # self.fig.canvas.draw()
            # self.fig.canvas.flush_events()
            # while self.fig.paused:
            #    self.fig.canvas.flush_events()

    def __getstate__(self):
        state = copy(self.__dict__)
        del state["mng"]

        del state["attachee"]

        if "canvas" in state:
            del state["canvas"]

        # del state["log"]
        # del state["exception"]
        # del state["mng"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.mng = Mock()
        self.attachee = Mock()

    def reset(self):
        self.frame_data = []
        if self.interactive_mode:
            # self.mng.window.close()
            pass
        self.ax.clear()
        self.setup()
        # self.__init__(attachee=self.attachee, interactive=self.interactive_mode)

    @abstractmethod
    def construct_frame(self, **frame_datum):
        pass

    def animate(self, frames=None):
        temp_fig, temp_ax = self.fig, self.ax
        plt.ioff()
        self.fig = Figure(figsize=(10, 10))
        canvas = FigureCanvas(self.fig)
        self.ax = canvas.figure.add_subplot(111)
        if frames is None:
            frames = self.__class__._frames
        if frames == "all":
            frames = len(self.frame_data)
        elif isinstance(frames, int):
            frames = max(min(frames, len(self.frame_data)), 0)
        elif isinstance(frames, float) and 0 <= frames <= 1:
            frames = int(frames * len(self.frame_data))
        else:
            raise ValueError(
                'animate accepts an int, a float or "all", but instead a different value was provided.'
            )
        if not hasattr(self, "trajectory"):

            def animation_update(i):
                j = int((i / frames) * len(self.frame_data) + 0.5)
                self.lim(frame_idx=j)
                return self.construct_frame(**self.frame_data[j])

        else:
            trajectory_xs = self.trajectory_xs
            trajectory_ys = self.trajectory_ys

            def animation_update(i):
                j = int((i / frames) * len(self.frame_data) + 0.5)
                self.lim(frame_idx=j)
                return self.construct_frame(
                    trajectory_xs=trajectory_xs[:j],
                    trajectory_ys=trajectory_ys[:j],
                    **self.frame_data[j],
                )

        self.setup()

        anim = matplotlib.animation.FuncAnimation(
            self.fig,
            animation_update,
            frames=frames,
            interval=30,
            blit=True,
            repeat=False,
            #init_func=self.setup
        )
        res = anim.to_jshtml()
        plt.close(self.fig)
        self.fig, self.ax = temp_fig, temp_ax
        return res

    def on_launch(self):
        os.mkdir(self.get_save_directory())

    def animate_and_save(self, frames=None, name=None):
        if name is None:
            name = str(self.saved_counter)
            self.saved_counter += 1
        animation = self.animate(frames=frames)
        dir = self.get_save_directory()
        with open(dir + "/" + name + ".html", "w") as f:
            f.write(
                f"<html><head><title>{self.__class__.__name__}: {name}</title></head><body>{animation}</body></html>"
            )

    @classmethod
    def lim_from_reference(cls, x, y, extra_margin=0, equal=True):
        x = np.array(x)
        y = np.array(y)
        if x.size == 0:
            x = np.zeros(1)
        if y.size == 0:
            y = np.zeros(1)
        x_min, x_max = x.min(), x.max()
        y_min, y_max = y.min(), y.max()
        x_min, x_max = x_min - (x_max - x_min) * 0.1, x_max + (x_max - x_min) * 0.1
        y_min, y_max = y_min - (y_max - y_min) * 0.1, y_max + (y_max - y_min) * 0.1
        left = x_min - extra_margin
        width = max(x_max - x_min, y_max - y_min) if equal else x_max - x_min
        right = x_min + width + extra_margin
        bottom = y_min - extra_margin
        height = max(x_max - x_min, y_max - y_min) if equal else y_max - y_min
        top = y_min + height + extra_margin
        return left, right, bottom, top

    def lim(self, width=None, height=None, center=None, extra_margin=0.0, **kwargs):
        if width is not None or height is not None:
            if center is None:
                center = [0.0, 0.0]
            if width is None:
                width = height
            if height is None:
                height = width
            self.ax.set_xlim(
                center[0] - width / 2 - extra_margin,
                center[0] + width / 2 + extra_margin,
            )
            self.ax.set_ylim(
                center[1] - height / 2 - extra_margin,
                center[1] - height / 2 + extra_margin,
            )
        else:
            raise ValueError("No axis limits known for animation.")

    def on_episode_done(
        self,
        scenario,
        episode_number,
        episodes_total,
        iteration_number,
        iterations_total,
    ):
        self.counter = 0
        # self.animate_and_save(name=str(episode_number))
        self.reset()

    def install_temp_axis(self, fig, ax, interactive=None):
        self._old_ax = self.ax
        self._old_fig = self.fig
        self._old_interactive_mode = self.interactive_mode
        self.fig = fig
        self.ax = ax
        if isinstance(self.ax, SubplotSpec):
            self.ax = fig.add_subplot(self.ax)
        if interactive is not None:
            self.interactive_mode = interactive

    def pop_axis(self):
        self.fig = self._old_fig
        self.ax = self._old_ax
        self.interactive_mode = self._old_interactive_mode
        del self._old_fig
        del self._old_ax
        del self._old_interactive_mode

    @property
    def latest_frame(self):
        return self.frame_data[-1]

    @property
    def latest_frame_idx(self):
        return len(self.frame_data) - 1


class PausableFigure(Figure):
    def __init__(self, *args, log=print, **kwargs):
        super().__init__(*args, **kwargs)
        self.paused = False

        def on_press(event):
            if event.key == " ":
                self.paused = not self.paused
            log("Paused." if self.paused else "Resumed.")

        self.canvas.mpl_connect("key_press_event", on_press)


def title_except_units(s):
    words = s[:s.find('[')]
    units = s[s.find('['):]
    return words.title() + units

class ComposedAnimationCallback(AnimationCallback):
    """An animation callback capable of incoroporating several other animation callbacks in such a way that the respective plots are distributed between axes of a single figure."""

    def __init__(
        self, *animations, fig=None, ax=None, mng=None, mode="square", **kwargs
    ):
        """Initialize an instance of ComposedAnimationCallback.

        Args:
            *animations (*List[AnimationCallback]): animation classes to be composed
            **kwargs (**Dict[str, Any]): keyword arguments to be passed to __init__ of said
                animations and the base class
        """
        callback.Callback.__init__(self, **kwargs)

        self.show_cooldown = 1 / float(self._metadata["argv"].fps)

        self.frame_data = []
        self.frame_indices = []

        self.interactive_mode = self._metadata["argv"].interactive

        self.paused = False

        # canvas = FigureCanvas(self.fig)
        if mode == "square":
            self.width = int(len(animations) ** 0.5 + 0.5)
            self.height = self.width - (self.width**2 - len(animations)) // self.width
        if mode == "vstack":
            self.width = 1
            self.height = len(animations)
        elif mode == "hstack":
            self.height = 1
            self.width = len(animations)
        else:
            assert mode in ["square", "vstack", "hstack"]

        self.figsize = (self.height * 5, self.width * 5)

        if fig is None and self.interactive_mode:
            self.fig = PausableFigure(log=self.log, figsize=self.figsize)
            self.mng = backend_qt5agg.new_figure_manager_given_figure(1, self.fig)
            win = self.fig.canvas.window()
            win.setFixedSize(win.size())
        elif fig is None:
            self.fig = Figure(figsize=self.figsize)
            self.mng = None
        else:
            self.fig = fig
            self.mng = mng

        if ax is None:
            gs = GridSpec(self.width, self.height, figure=self.fig)
        else:
            gs = GridSpecFromSubplotSpec(self.width, self.height, subplot_spec=ax)

        self.animations = []
        for i, animation in enumerate(animations):
            self.animations.append(
                animation(
                    interactive=False, fig=self.fig, ax=gs[i], mng=self.mng, **kwargs
                )
            )

        # self.animations = [Animation(interactive=False, fig=self.fig, ax=gs, **kwargs) for Animation in animations]

        self.save_directory = Path(
            f".callbacks/{self.__class__.__name__}"
            + (f"@{self.attachee.__name__}" if self.attachee is not None else "")
        ).resolve()

        for animation in self.animations:
            animation.interactive_mode = self._metadata["argv"].interactive
            animation.setup()
        if self.interactive_mode:
            self.fig.show()

        self.ax = None
        self.last_show = time.time()
        # plt.show(block=True)

    @property
    def latest_frame(self):
        return [animation.latest_frame for animation in self.animations]

    @property
    def latest_frame_idx(self):
        return [animation.latest_frame_idx for animation in self.animations]

    def on_function_call(self, obj, method, output):
        try:
            self.frame_data.append(self.latest_frame)
            self.frame_indices.append(self.latest_frame_idx)
        except (IndexError, PrematureResolutionError):
            pass
        if self.interactive_mode:
            present_show = time.time()
            if present_show - self.last_show > self.show_cooldown:
                self.fig.canvas.draw()
                self.last_show = present_show
            self.fig.canvas.flush_events()
            while self.fig.paused:
                self.fig.canvas.flush_events()

    def construct_frame(self, *frame_data_of_animations):
        actor_bundles = []
        for animation, datum in zip(self.animations, frame_data_of_animations):
            actor_bundle = (
                animation.construct_frame(**datum)
                if not isinstance(animation, ComposedAnimationCallback)
                else animation.construct_frame(*datum)
            )
            actor_bundles.append(actor_bundle)
        return [actor for actors in actor_bundles for actor in actors]

    def setup(self):
        for animation in self.animations:
            animation.setup()

    def is_target_event(self, obj, method, output, triggers):
        return True # This is too much

    def on_launch(self):
        os.mkdir(self.get_save_directory())
        for animation in self.animations:
            animation.on_launch()

    def on_episode_done(
        self,
        scenario,
        episode_number,
        episodes_total,
        iteration_number,
        iterations_total,
    ):
        playback = (
            self._metadata["argv"].playback
            and iteration_number == iterations_total
            and episode_number == episodes_total
        )
        save_animation = self._metadata["argv"].save_animation
        if playback or save_animation:
            self.animate_and_save(
                name=f"ep{episode_number}|{episodes_total},it{iteration_number}|{iterations_total}"
            )
        for animation in self.animations:
            animation.on_episode_done(
                scenario,
                episode_number,
                episodes_total,
                iteration_number,
                iterations_total,
            )
        self.frame_data = []

    def on_iteration_done(self, *args, **kwargs):
        for animation in self.animations:
            animation.on_iteration_done(*args, **kwargs)

    def __del__(self):
        for animation in self.animations:
            del animation

    def __call__(self, *args, **kwargs):
        super().__call__(*args, **kwargs)
        for animation in self.animations:
            animation(*args, **kwargs)

    def lim(self, *args, frame_idx=None, **kwargs):
        for animation, idx in zip(self.animations, frame_idx):
            animation.lim(*args, frame_idx=idx, **kwargs)

    def animate(self, frames=None):
        plt.ioff()
        fig = Figure(figsize=self.figsize)
        canvas = FigureCanvas(fig)
        gs = GridSpec(self.width, self.height, figure=self.fig)
        self.install_temp_axis(fig, gs, interactive=False)
        if frames is None:
            frames = self.__class__._frames
        if frames == "all":
            frames = len(self.frame_data)
        elif isinstance(frames, int):
            frames = max(min(frames, len(self.frame_data)), 0)
        elif isinstance(frames, float) and 0 <= frames <= 1:
            frames = int(frames * len(self.frame_data))
        else:
            raise ValueError(
                'animate accepts an int, a float or "all", but instead a different value was provided.'
            )
        if not hasattr(self, "trajectory"):

            def animation_update(i):
                j = int((i / frames) * len(self.frame_data) + 0.5)
                self.lim(frame_idx=self.frame_indices[j])
                return self.construct_frame(*self.frame_data[j])

        else:
            trajectory_xs = self.trajectory_xs
            trajectory_ys = self.trajectory_ys

            def animation_update(i):
                j = int((i / frames) * len(self.frame_data) + 0.5)
                self.lim(frame_idx=j)
                return self.construct_frame(
                    trajectory_xs=trajectory_xs[:j],
                    trajectory_ys=trajectory_ys[:j],
                    *self.frame_data[j],
                )

        self.setup()

        anim = matplotlib.animation.FuncAnimation(
            self.fig,
            animation_update,
            frames=frames,
            interval=30,
            blit=True,
            repeat=False,
        )
        res = anim.to_jshtml()
        plt.close(self.fig)

        self.pop_axis()

        return res

    def install_temp_axis(self, fig, gs, interactive=None):
        if self.ax is not None:
            gs = GridSpecFromSubplotSpec(self.height, self.width, subplot_spec=gs)
        self._old_fig = self.fig
        self.fig = fig
        for i, animation in enumerate(self.animations):
            animation.install_temp_axis(fig, gs[i], interactive=interactive)

    def pop_axis(self):
        self.fig = self._old_fig
        del self._old_fig
        for animation in self.animations:
            animation.pop_axis()


class PrematureResolutionError(Exception):
    pass


class DeferredComposedAnimation(ComposedAnimationCallback, ABC):
    cooldown = None

    def __init__(
        self, *animations, fig=None, ax=None, mode="vstack", mng=None, **kwargs
    ):
        self.initialized = False
        if "interactive" in kwargs:
            del kwargs["interactive"]
        callback.Callback.__init__(self, **kwargs)

        """
        self.fig = Figure(figsize=(10, 10))
        self.canvas = FigureCanvas(self.fig)
        self._animation_classes = animations
        self._kwargs = kwargs
        """

        if fig is None:
            self.fig = Figure(figsize=(10, 10))
            self.mng = backend_qt5agg.new_figure_manager_given_figure(1, self.fig)
        else:
            self.fig = fig
            self.mng = mng

        self.ax = ax
        self.mode = mode
        self._animation_classes = animations
        # canvas = FigureCanvas(self.fig)

        self._kwargs = kwargs

        self.save_directory = Path(
            f".callbacks/{self.__class__.__name__}"
            + (f"@{self.attachee.__name__}" if self.attachee is not None else "")
        ).resolve()

    def on_launch(self):
        if self.initialized:
            super().on_launch()

    @property
    def animations(self):
        if self.initialized:
            return self._animations
        else:
            raise PrematureResolutionError(
                f"Tried to access attribute 'animation' of class {self.__class__.__name__} before the instance initialized."
            )

    def on_function_call(self, obj, method, output):
        pass

    def __deferred_init__(self):
        """
        self.animations = [
            Animation(interactive=False, **self._kwargs) for Animation in self._animation_classes
        ]
        width = int(len(self.animations) ** 0.5 + 0.5)
        height = width - (width**2 - len(self.animations)) // width
        for i, animation in enumerate(self.animations):
            animation.fig = self.fig
            animation.ax = self.canvas.figure.add_subplot(width, height, 1 + i)
        self.mng = backend_qt5agg.new_figure_manager_given_figure(1, self.fig)
        for animation in self.animations:
            animation.mng = self.mng
            animation.interactive_mode = self._metadata["argv"].interactive
            animation.setup()
            animation.fig.show()
        """

        if self.mode == "square":
            self.width = int(len(self._animation_classes) ** 0.5 + 0.5)
            self.height = (
                self.width
                - (self.width**2 - len(self._animation_classes)) // self.width
            )
        if self.mode == "vstack":
            self.width = 1
            self.height = len(self._animation_classes)
        elif self.mode == "hstack":
            self.height = 1
            self.width = len(self._animation_classes)
        else:
            assert self.mode in ["square", "vstack", "hstack"]

        if self.ax is None:
            self.gs = GridSpec(self.height, self.width, figure=self.fig)
        else:
            self.gs = GridSpecFromSubplotSpec(
                self.height, self.width, subplot_spec=self.ax
            )

        self._animations = []
        for i, animation in enumerate(self._animation_classes):
            self._animations.append(
                animation(
                    interactive=False,
                    fig=self.fig,
                    ax=self.gs[i],
                    mng=self.mng,
                    **self._kwargs,
                )
            )

        # self.animations = [Animation(interactive=False, fig=self.fig, ax=gs, **kwargs) for Animation in animations]

        for animation in self._animations:
            animation.interactive_mode = self._metadata["argv"].interactive
            animation.setup()
        # self.fig.show()

        self.initialized = True
        del self._animation_classes
        del self._kwargs

    def __call__(self, *args, **kwargs):
        if self.initialized:
            for animation in self.animations:
                animation(*args, **kwargs)
        else:
            callback.Callback.__call__(self, *args, **kwargs)

    def setup(self):
        if self.initialized:
            for animation in self.animations:
                animation.setup()

    def on_episode_done(self, *args, **kwargs):
        for animation in self.animations:
            animation.on_episode_done(*args, **kwargs)


class StateAnimation(DeferredComposedAnimation, callback.StateTracker):
    """A graph animation that plots values of state variables vs. time."""
    def __deferred_init__(self):
        state_dimension = len(self.system_state)
        self._animation_classes = []
        for i in range(state_dimension):

            def stateComponent(*args, component=i, **kwargs):
                return StateComponentAnimation(*args, component=component, fontsize=min(10.0, 15.0/state_dimension**0.50), **kwargs)

            self._animation_classes.append(stateComponent)
        super().__deferred_init__()

    def is_target_event(self, obj, method, output, triggers):
        return callback.StateTracker in triggers

    def on_trigger(self, _):
        self.__deferred_init__()
        super().on_launch()  # a bit clumsy, I think


class ObservationAnimation(DeferredComposedAnimation, callback.ObservationTracker):
    """A graph animation that plots observables vs. time."""
    def __deferred_init__(self):
        observation_dimension = len(self.observation)
        self._animation_classes = []
        for i in range(observation_dimension):

            def observationComponent(*args, component=i, **kwargs):
                return ObservationComponentAnimation(
                    *args, component=component, fontsize=min(10.0, 15.0/observation_dimension**0.5), **kwargs
                )

            self._animation_classes.append(observationComponent)
        super().__deferred_init__()

    def is_target_event(self, obj, method, output, triggers):
        return callback.ObservationTracker in triggers

    def on_trigger(self, _):
        self.__deferred_init__()
        super().on_launch()  # a bit clumsy, I think


class ActionAnimation(DeferredComposedAnimation, callback.ActionTracker):
    """A graph animation that plots control inputs vs. time."""
    def __deferred_init__(self):
        action_dimension = len(self.action)
        self._animation_classes = []
        for i in range(action_dimension):

            def actionComponent(*args, component=i, **kwargs):
                return ActionComponentAnimation(*args, component=component, fontsize=min(10.0, 15.0/action_dimension**0.5), **kwargs)

            self._animation_classes.append(actionComponent)
        super().__deferred_init__()

    def is_target_event(self, obj, method, output, triggers):
        return callback.ActionTracker in triggers

    def on_trigger(self, _):
        self.__deferred_init__()
        super().on_launch()  # a bit clumsy, I think


class ObjectiveAnimation(DeferredComposedAnimation, callback.ObjectiveTracker):
    def __deferred_init__(self):
        objective_dimension = len(self.objective)
        self._animation_classes = []
        for i in range(objective_dimension):

            def objectiveComponent(*args, component=i, **kwargs):
                return ObjectiveComponentAnimation(*args, component=component, fontsize=min(10.0, 15.0/objective_dimension**0.5), **kwargs)

            self._animation_classes.append(objectiveComponent)
        super().__deferred_init__()

    def is_target_event(self, obj, method, output, triggers):
        return callback.ObjectiveTracker in triggers

    def on_trigger(self, _):
        self.__deferred_init__()
        super().on_launch()


class GraphAnimation(AnimationCallback):
    """Animation of graphs that adds more data to curves over time."""
    _legend = (None,)
    _vertices = 4000
    _line = "-"

    def add_frame(self, line_datas):
        line_datas = list(line_datas)
        for i in range(len(line_datas)):
            t, y = line_datas[i]
            line_datas[i] = (list(t), list(y))
        super().add_frame(line_datas=line_datas)

    def setup(self):
        self.lines = []
        for name in self._legend:
            (line,) = self.ax.plot([0], [0], self._line, label=name)
            self.lines.append(line)

    def lim(self, *args, extra_margin=0.01, **kwargs):
        datum = self.frame_data[-1]["line_datas"]
        ts, ys = [], []
        for line_data in datum:
            t, y = line_data
            skip = 1 + len(t) // self._vertices
            ts += list(t)[::skip]
            ys += list(y)[::skip]

        left, right, bottom, top = self.lim_from_reference(
            np.array(ts), np.array(ys), extra_margin, equal=False
        )
        self.ax.set_xlim(left, right)
        self.ax.set_ylim(bottom, top)

    def construct_frame(self, line_datas):
        for line, data in zip(self.lines, line_datas):
            t, y = data
            skip = 1 + len(t) // self._vertices
            downsampling = slice(1, None, skip) if isinstance(self, callback.TimeTracker) else slice(None, None, skip)
            line.set_data(t[downsampling], y[downsampling]) # these slices are to avoid residual time bug
        return self.lines


class ValueAnimation(GraphAnimation, callback.ValueTracker):
    """A graph animation that displays that plots episode-mean value over iterations."""
    _legend = (None,)
    _is_global = True

    def setup(self):
        super().setup()
        if not hasattr(self, "scores"):
            self.t = []
            self.y = []
            self.scores = []
            self.ax.set_axis_off()
        self.ax.set_xlabel("Learning Iteration")
        self.ax.set_ylabel("Value")
        self.ax.xaxis.set_major_locator(MaxNLocator(integer=True))
        self.add_frame(line_datas=[(self.t, self.y)])

    def install_temp_axis(self, *args, **kwargs):
        super().install_temp_axis(*args, **kwargs)
        self.lines_old = self.lines

    def pop_axis(self):
        super().pop_axis()
        self.lines = self.lines_old
        del self.lines_old

    def is_target_event(self, obj, method, output, triggers):
        return callback.ValueTracker in triggers

    def on_trigger(self, _):
        self.scores.append(self.score)

    def on_iteration_done(
        self,
        scenario,
        episode_number,
        episodes_total,
        iteration_number,
        iterations_total,
    ):
        if iteration_number >= 2:
            self.ax.set_axis_on()
        if self.scores:
            self.t.append(iteration_number)
            self.y.append(np.mean(self.scores))
            self.add_frame(
                line_datas=[(self.t, self.y)]
            )  # these slices are there to avoid residual time bug
        self.scores = []

    def on_episode_done(
        self,
        scenario,
        episode_number,
        episodes_total,
        iteration_number,
        iterations_total,
    ):
        pass #self.setup()


class StateComponentAnimation(
    GraphAnimation, callback.StateTracker, callback.TimeTracker
):
    _legend = (None,)

    def __init__(self, *args, component=0, fontsize=10.0, **kwargs):
        super().__init__(*args, **kwargs)
        self.component = component
        self.fontsize = fontsize

    def get_save_directory(self):
        return super().get_save_directory() + str(self.component)

    def setup(self):
        super().setup()
        self.ax.set_xlabel("Time [s]")
        if hasattr(self, "label_of_y"):
            self.ax.set_ylabel(self.label_of_y, fontsize=self.fontsize)
        self.t = []
        self.y = []

    def is_target_event(self, obj, method, output, triggers):
        return callback.StateTracker in triggers

    def on_trigger(self, _):
        if not hasattr(self, "label_of_y"):
            self.label_of_y = title_except_units(self.state_naming[self.component])
            self.ax.set_ylabel(self.label_of_y, fontsize=self.fontsize)
        self.t.append(self.time)
        self.y.append(self.system_state[self.component])
        self.add_frame(
            line_datas=[(self.t, self.y)]
        )  # these slices are there to avoid residual time bug


class ObservationComponentAnimation(  # TO DO: introduce an abstract class ObservablesAnimation which will include its own ObservableComponentAnimation.
    GraphAnimation, callback.ObservationTracker, callback.TimeTracker
):
    _legend = (None,)

    def __init__(self, *args, component=0, fontsize=10.0, **kwargs):
        super().__init__(*args, **kwargs)
        self.component = component
        self.fontsize = fontsize

    def get_save_directory(self):
        return super().get_save_directory() + str(self.component)

    def setup(self):
        super().setup()
        self.ax.set_xlabel("Time [s]")
        if hasattr(self, "label_of_y"):
            self.ax.set_ylabel(self.label_of_y, fontsize=self.fontsize)
        self.t = []
        self.y = []

    def is_target_event(self, obj, method, output, triggers):
        return callback.ObservationTracker in triggers

    def on_trigger(self, _):
        if not hasattr(self, "label_of_y"):
            self.label_of_y = title_except_units(self.observation_naming[self.component])
            self.ax.set_ylabel(self.label_of_y, fontsize=self.fontsize)
        self.t.append(self.time)
        self.y.append(self.observation[self.component])
        self.add_frame(
            line_datas=[(self.t, self.y)]
        )  # these slices are there to avoid residual time bug


class ActionComponentAnimation(  # TO DO: introduce an abstract class ObservablesAnimation which will include its own ObservableComponentAnimation.
    GraphAnimation, callback.ActionTracker, callback.TimeTracker
):
    _legend = (None,)
    _line = "g-"

    def __init__(self, *args, component=0, fontsize=10.0, **kwargs):
        super().__init__(*args, **kwargs)
        self.component = component
        self.fontsize = fontsize

    def get_save_directory(self):
        return super().get_save_directory() + str(self.component)

    def setup(self):
        super().setup()
        self.ax.set_xlabel("Time [s]")
        if hasattr(self, "label_of_y"):
            self.ax.set_ylabel(self.label_of_y, fontsize=self.fontsize)
        self.t = []
        self.y = []

    def is_target_event(self, obj, method, output, triggers):
        return callback.ActionTracker in triggers

    def on_trigger(self, _):
        if not hasattr(self, "label_of_y"):
            self.label_of_y = title_except_units(self.action_naming[self.component])
            self.ax.set_ylabel(self.label_of_y, fontsize=self.fontsize)
        self.t.append(self.time)
        self.y.append(self.action[self.component])
        self.add_frame(
            line_datas=[(self.t, self.y)]
        )  # these slices are there to avoid residual time bug


class ObjectiveComponentAnimation(  # TO DO: introduce an abstract class ObservablesAnimation which will include its own ObservableComponentAnimation.
    GraphAnimation, callback.ObjectiveTracker, callback.TimeTracker
):
    _legend = (None,)
    _line = "r-"

    def __init__(self, *args, component=0, fontsize=10.0, **kwargs):
        super().__init__(*args, **kwargs)
        self.component = component
        self.fontsize = fontsize

    def get_save_directory(self):
        return super().get_save_directory() + str(self.component)

    def setup(self):
        super().setup()
        self.ax.set_xlabel("Time [s]")
        if hasattr(self, "label_of_y"):
            self.ax.set_ylabel(self.label_of_y, fontsize=self.fontsize)
        self.t = []
        self.y = []

    def is_target_event(self, obj, method, output, triggers):
        return callback.ObjectiveTracker in triggers

    def on_trigger(self, _):
        if not hasattr(self, "label_of_y"):
            self.label_of_y = title_except_units(self.objective_naming[self.component])
            self.ax.set_ylabel(self.label_of_y, fontsize=self.fontsize)
        self.t.append(self.time)
        self.y.append(self.objective[self.component])
        self.add_frame(
            line_datas=[(self.t, self.y)]
        )  # these slices are there to avoid residual time bug


class MultiSpriteAnimation(AnimationCallback, ABC):
    """Animation that sets the location and rotation of a planar shape at each frame."""

    _pics = None  # must be svgs located in regelum/img
    _marker_sizes = 30
    _rotations = 0
    _center_vert = False

    def setup(self):
        self.trajectory_xs = []
        self.trajectory_ys = []
        (self.trajectory,) = self.ax.plot(self.trajectory_xs, self.trajectory_ys, "--")
        self.paths = [
            regelum.__file__.replace("__init__.py", f"img/{pic}") for pic in self._pics
        ]
        self.markers = []
        self.sprites = []
        self.pic_datas = []
        self.attributes = []
        self.original_transforms = []
        if not isinstance(self._rotations, Iterable):
            rotations = [self._rotations] * len(self.paths)
        else:
            rotations = self._rotations
        if not isinstance(self._marker_sizes, Iterable):
            sizes = [self._marker_sizes] * len(self.paths)
        else:
            sizes = self._marker_sizes
        for path, rotation, size in zip(self.paths, rotations, sizes):
            pic_data, attribute = svg2paths(path)
            parsed = parse_path(attribute[0]["d"])
            parsed.vertices[:, 0] -= parsed.vertices[:, 0].mean(axis=0)
            if self._center_vert:
                parsed.vertices[:, 1] -= parsed.vertices[:, 1].mean(axis=0)
            marker = matplotlib.markers.MarkerStyle(marker=parsed)
            marker._transform = marker.get_transform().rotate_deg(rotation)
            (sprite,) = self.ax.plot(0, 1, marker=marker, ms=size)
            original_transform = marker.get_transform()
            self.markers.append(marker)
            self.sprites.append(sprite)
            self.pic_datas.append(pic_data)
            self.attributes.append(attribute)
            self.original_transforms.append(original_transform)

    def increment_trajectory(self, x0=None, y0=None, **frame_datum):
        return x0, y0

    def lim(self, *args, extra_margin=0.11, **kwargs):
        x, y = np.array([list(datum.values()) for datum in self.frame_data]).T[:2]
        left, right, bottom, top = self.lim_from_reference(x, y, extra_margin)
        self.ax.set_xlim(left, right)
        self.ax.set_ylim(bottom, top)

    def add_frame(self, **frame_datum):
        traj_x, traj_y = self.increment_trajectory(**frame_datum)
        self.trajectory_xs.append(traj_x)
        self.trajectory_ys.append(traj_y)
        frame_datum["trajectory_data"] = (
            list(self.trajectory_xs),
            list(self.trajectory_ys),
        )
        return super().add_frame(**frame_datum)

    def construct_frame(self, trajectory_data=None, **kwargs):
        self.trajectory.set_data(*trajectory_data)
        for i in range(len(self.paths)):
            x, y, theta = kwargs[f"x{i}"], kwargs[f"y{i}"], kwargs[f"theta{i}"]
            self.sprites[i].set_data([x], [y])
            self.markers[i]._transform = Affine2D(
                self.original_transforms[i]._mtx.copy()
            )
            self.markers[i]._transform = (
                self.markers[i].get_transform().rotate_deg(180 * theta / np.pi)
            )
        return self.sprites


class PlanarBodiesAnimation(MultiSpriteAnimation, callback.StateTracker):
    """Animates dynamics of systems that can be viewed as a set of bodies moving on a plane."""

    def setup(self):
        super().setup()
        self.ax.set_xlabel("X [m]")
        self.ax.set_ylabel("Y [m]")
        # self.ax.set_title(f"Planar motion of {self.attachee.__name__}")

    def on_trigger(self, _):
        self.add_frame(
            x0=self.system_state[0],
            y0=self.system_state[1],
            theta0=self.system_state[2],
        )


class CartpoleAnimation(PlanarBodiesAnimation):
    """Animates dynamics of Lunar Lander that can be viewed as a triangle moving on a plane."""

    _pics = ["cart.svg", "pendulum.svg"]
    _marker_sizes = [60, 100]

    def lim(self, *args, frame_idx=None, extra_margin=0.11, **kwargs):
        x = self.frame_data[frame_idx]["x0"]
        left, right, bottom, top = self.lim_from_reference(
            np.array([x - 2, x + 2]), np.array([-1, 3]), extra_margin, equal=True
        )
        self.ax.set_xlim(left, right)
        self.ax.set_ylim(bottom, top)

    def increment_trajectory(self, x0, theta1, **kwargs):
        return x0 + np.cos(theta1 + np.pi / 2), np.sin(theta1 + np.pi / 2)

    def setup(self):
        super().setup()
        self.ax.set_xlabel("X [m]")
        self.ax.set_ylabel("")
        self.ax.get_yaxis().set_visible(False)
        self.ax.set_aspect("equal", adjustable="box")
        (self.ground,) = self.ax.plot([-100000, 100000], [0, 0], "r", ms=10, zorder=-1)
        (self.target,) = self.ax.plot([0], [0], "g", ms=35, marker="o", zorder=-1)
        self.target_text = self.ax.text(
            0,
            0,
            "Target",
            horizontalalignment="center",
            verticalalignment="center",
            zorder=-1,
        ).set_clip_on(True)

    def on_trigger(self, _):
        self.add_frame(
            x0=self.system_state[1],
            y0=0,
            theta0=0,
            x1=self.system_state[1],
            y1=0,
            theta1=self.system_state[0],
        )


class SingleBodyAnimation(PlanarBodiesAnimation):
    """Animates dynamics of systems that can be viewed as a body moving on a plane."""

    _pic = "default.svg"
    _marker_size = None
    _rotation = None

    def setup(self):
        self._pics = [self._pic]
        if self._marker_size is not None:
            self._marker_sizes = [self._marker_size]
        if self._rotation is not None:
            self._rotations = [self._rotation]
        super().setup()

    def add_frame(self, **frame_datum):
        if "x" in frame_datum:
            frame_datum["x0"] = frame_datum["x"]
        if "y" in frame_datum:
            frame_datum["y0"] = frame_datum["y"]
        if "theta" in frame_datum:
            frame_datum["theta0"] = frame_datum["theta"]
        super().add_frame(**frame_datum)


class LunarLanderAnimation(SingleBodyAnimation):
    """Animates dynamics of Lunar Lander that can be viewed as a triangle moving on a plane."""

    _pic = "lunar_lander.svg"
    _marker_size = 30
    _center_vert = True

    def lim(self, *args, frame_idx=None, extra_margin=0.11, **kwargs):
        self.sprites[0].set_visible(True)

        x = self.frame_data[frame_idx]["x"]
        left, right, bottom, top = self.lim_from_reference(
            np.array([x - 2, x + 2]), np.array([0, 5]), extra_margin, equal=False
        )
        self.ax.set_xlim(left, right)
        self.ax.set_ylim(bottom, top)

    def setup(self):
        super().setup()
        #(self.ground,) = self.ax.plot([-100000, 100000], [0, 0], "r", ms=10)
        self.ax.set_ylim(0 - 0.11, 5 + 0.11)
        self.ax.set_xlim(-2 - 0.11, 2 + 0.11)
        self.sprites[0].set_visible(False)
        rect = Rectangle((-10000, -1), 20000, 1, linewidth=1, edgecolor='grey', facecolor='grey')
        self.ground = self.ax.add_patch(rect)
        (self.target,) = self.ax.plot([0], [0], "g", ms=35, marker="o")
        self.target_text = self.ax.text(
            0, 0, "Target", horizontalalignment="center", verticalalignment="center"
        ).set_clip_on(True)

    def on_trigger(self, _):
        self.add_frame(
            x=self.system_state[2], y=self.system_state[1], theta=self.system_state[0]
        )


class OmnirobotAnimation(SingleBodyAnimation):
    _pic = "omnirobot.svg"
    _marker_size = 30
    _center_vert = True

    def on_trigger(self, _):
        self.add_frame(x=self.system_state[0], y=self.system_state[1], theta=0)

    def lim(self, *args, **kwargs):
        self.ax.set_xlim(-15, 15)
        self.ax.set_ylim(-15, 15)


class PendulumAnimation(SingleBodyAnimation):
    """Animates the head of a swinging pendulum.

    Interprets the first state coordinate as the angle of the pendulum with respect to the topmost position.
    """

    _pic = "pendulum.svg"
    _marker_size = 250

    def setup(self):
        super().setup()
        self.ax.get_yaxis().set_visible(False)
        self.ax.get_xaxis().set_visible(False)
        self.ax.set_aspect("equal", adjustable="box")
        (self.trajectory,) = self.ax.plot(
            self.trajectory_xs, self.trajectory_ys, "--", alpha=0.6
        )

    def on_trigger(self, _):
        self.add_frame(x=0, y=0, theta=self.system_state[0])

    def increment_trajectory(self, x, y, theta, **kwargs):
        return np.cos(theta + np.pi / 2) * 0.9, np.sin(theta + np.pi / 2) * 0.9

    def lim(self, *args, **kwargs):
        self.ax.set_xlim(-1, 1)
        self.ax.set_ylim(-1, 1)


class ThreeWheeledRobotAnimation(SingleBodyAnimation):
    """Animates the position of a differential platform."""

    _pic = "3wrobot.svg"
    _rotation = 225
    # _marker_size = 250
    # _frames = 500

    def lim(self, *args, **kwargs):
        self.ax.set_xlim(-7, 7)
        self.ax.set_ylim(-7, 7)


class BarAnimation(AnimationCallback, callback.StateTracker):
    """Animates the state of a system as a collection of vertical bars."""

    _naming = None

    def setup(self):
        self.bar = None

    def construct_frame(self, state):
        if self.bar is None:
            self.bar = self.ax.bar(self.state_naming, state)
        else:
            for rect, h in zip(self.bar, state):
                rect.set_height(h)
        return self.bar

    def on_trigger(self, _):
        self.add_frame(state=self.system_state)

    def lim(self, *args, **kwargs):
        pass


@contextmanager
def preserve_limits(ax=None):
    """ Plot without modifying axis limits """

    xlim = ax.get_xlim()
    ylim = ax.get_ylim()

    try:
        yield ax
    finally:
        ax.set_xlim(xlim)
        ax.set_ylim(ylim)

class TwoTankAnimation(BarAnimation):
    def setup(self):
        super().setup()
        self.level = self.ax.plot([-10000, 10000], [0.44, 0.44], 'r--')
        pump = Rectangle((0.4, 0), 0.2, 0.1, zorder=-1, facecolor="grey")
        self.ax.add_patch(pump)
        self.ax.text(0.4, 0.01, "Pump", fontsize=8, zorder=11)

        sink = Rectangle((1.2, 0), 0.4, 0.1, zorder=10, facecolor="grey")
        self.ax.add_patch(sink)
        self.ax.text(1.2, 0.01, "Sink", fontsize=8, zorder=11)

        intake = Rectangle((-0.5, 1.8), 0.2, 0.1, zorder=10, facecolor="grey")
        self.ax.add_patch(intake)
        self.ax.text(-0.5, 1.81, "Intake", fontsize=8, zorder=11)

        self.ax.set_xlim(-0.5, 1.5)



DefaultAnimation = StateAnimation + ActionAnimation + ValueAnimation



"""Tools for systems' state estimation."""

import numpy as np
from abc import ABC, abstractmethod

import regelum
from typing import Union, List


class Observer(regelum.RegelumBase, ABC):
    """A class implementing observer."""

    @abstractmethod
    def get_state_estimation(self): ...


class ObserverTrivial(Observer):
    """A class implementing a trivial observer, which simply returns the current state."""

    def get_state_estimation(self, t, observation, action):
        return observation


class ObserverReference(Observer):
    """Estimates state via adding reference to observation."""

    def __init__(self, reference: Union[np.ndarray, List[float]]):
        """Instatiate ObserverReference.

        Args:
            reference (Union[np.ndarray, List[float]]): array for
                reference
        """
        self.reference = np.array(reference).reshape(1, -1)

    def get_state_estimation(self, t, observation, action):
        return observation + self.reference


class CartPoleObserverPG(Observer):
    def __init__(self):
        pass

    def get_state_estimation(self, t, observation: np.ndarray, action):
        # sin_theta, one_minus_cos_theta, x, theta_dot, x_dot = observation.reshape(-1)
        sin_theta, one_minus_cos_theta, theta_dot, x_dot = observation.reshape(-1)

        return np.array(
            [[np.arctan2(sin_theta, 1 - one_minus_cos_theta), 0, theta_dot, x_dot]]
        )


class CartPoleObserver(Observer):
    def __init__(self):
        pass

    def get_state_estimation(self, t, observation: np.ndarray, action):
        sin_theta, one_minus_cos_theta, x, theta_dot, x_dot = observation.reshape(-1)
        # sin_theta, one_minus_cos_theta, theta_dot, x_dot = observation.reshape(-1)

        return np.array(
            [[np.arctan2(sin_theta, 1 - one_minus_cos_theta), x, theta_dot, x_dot]]
        )


class KalmanFilter(Observer):
    """A class implementing Kalman filter."""

    def __init__(
        self,
        t0,
        my_sys,
        sys_noise_cov,
        observ_noise_cov,
        prior_est_cov,
        state_init,
    ):
        """Initialize an instance of KalmanFilter.

        Args:
            t0: time at which simulation starts
            my_sys: an instance of a system of which state is to be
                observed
            sys_noise_cov: system noise covariance matrix
            observ_noise_cov: observation noise covariance matrix
            prior_est_cov: prior esimation covariance
            state_init: initial state
        """
        self.my_sys = my_sys

        self.posterior_state_est = state_init
        self.prior_state_est = None
        self.dim_state = self.posterior_state_est.shape[0]
        self.sys_noise_cov = sys_noise_cov
        self.observ_noise_cov = observ_noise_cov

        self.posterior_est_cov = np.eye(self.dim_state)
        self.prior_est_cov = prior_est_cov

        self.est_clock = t0

    def predict_state(self, action, dt):
        Q = self.sys_noise_cov
        J = self.my_sys.Jacobi_system_matrix(self.posterior_state_est)
        P_posterior_prev = self.posterior_est_cov

        self.prior_state_est = self.posterior_state_est + dt * self.my_sys._state_dyn(
            [], self.posterior_state_est, action
        )

        self.prior_est_cov = J @ P_posterior_prev @ J.T + Q

    def correct_state(self, observation):
        z = np.array(observation)
        J_h = self.my_sys.Jacobi_observation_matrix()
        P_pred = self.prior_est_cov
        R = self.observ_noise_cov

        K = np.array(P_pred @ J_h.T @ np.linalg.inv(J_h @ P_pred @ J_h.T + R))
        self.posterior_state_est = self.prior_state_est + K @ (
            z
            - self.my_sys.get_observation(
                time=None, state=self.prior_state_est, inputs=None
            )
        )
        self.posterior_est_cov = (np.eye(self.dim_state) - K @ J_h) @ P_pred
        print(f"Kalman gain:\n{K}, \n Post est cov: \n{self.posterior_est_cov}")

    def get_state_estimation(self, t, observation, action):
        dt = t - self.est_clock
        self.est_clock = t
        print(f"TRACE OF COV EST MATRIX:{self.prior_est_cov.trace()}\n")
        self.predict_state(action, dt)
        self.correct_state(observation)
        return self.posterior_state_est



"""Module contains policies, i.e., entities that directly calculate actions. 

The Policy calculates actions based on observations from 
the environment and the current state of the model. 
It can be optimized using various optimization techniques, 
adjusting the model's parameters to improve the chosen actions over time.

Policies are integrated into scenarios, allowing them to operate within agents 
that interact with dynamic environments, like those found in reinforcement learning.

Remarks:

- All vectors are treated as of type [n,]
- All buffers are treated as of type [L, n] where each row is a vector
- Buffers are updated from bottom to top

"""

import numpy as np
from abc import ABC, abstractmethod
from typing import Union, Optional
import casadi as cs
from .utils import rg, AwaitedParameter
from regelum.typing import Weights
from .predictor import Predictor
from .model import ModelNN, Model, ModelWeightContainer
from .critic import Critic, CriticTrivial
from .system import System, ComposedSystem
from .optimizable.optimizers import Optimizable, OptimizerConfig
from .data_buffers.data_buffer import DataBuffer
from .constraint_parser import ConstraintParser, ConstraintParserTrivial
from .objective import (
    RunningObjective,
    reinforce_objective,
    sdpg_objective,
    ddpg_objective,
    mpc_objective,
    rpv_objective,
    sql_objective,
    rql_objective,
    ppo_objective,
)
from typing import List, Tuple
from regelum.typing import RgArray
from typing import Any

import torch


class Policy(Optimizable, ABC):
    """Class of policies.

    Policy defines the strategy or rule by which an agent interacts with an environment.

    The Policy calculates actions based on observations from
    the environment and the current state of the model.
    It can be optimized using various optimization techniques,
    adjusting the model's parameters to improve the chosen actions over time.

    Policies are integrated into scenarios, allowing them to operate within agents
    that interact with dynamic environments, like those found in reinforcement learning.
    """

    def __init__(
        self,
        model: Union[Model, ModelNN] = None,
        system: Union[System, ComposedSystem] = None,
        action_bounds: Union[list, np.ndarray, None] = None,
        optimizer_config: Optional[OptimizerConfig] = None,
        discount_factor: Optional[float] = 1.0,
        epsilon_random_parameter: Optional[float] = None,
    ):
        """Initialize an instance of Policy class.

        Args:
            model: The model representing the policy's decision-making mechanism.
            system: System in which the policy will be deployed.
            action_bounds: Limits to the range of actions the policy can generate.
            optimizer_config: Configuration settings for the optimization procedure.
            discount_factor: Discount factor for the future running objectives.
            epsilon_random_parameter: Parameter that defines the randomness in action selection to encourage exploration.
        """
        self.system = system
        self.model = model

        self.dim_action = self.system.dim_inputs if self.system is not None else None
        self.dim_observation = (
            self.system.dim_observation if self.system is not None else None
        )

        self.discount_factor = discount_factor if discount_factor is not None else 1.0
        self.epsilon_random_parameter = epsilon_random_parameter
        if optimizer_config is not None:
            super().__init__(optimizer_config=optimizer_config)
        if self.dim_action is not None:
            (
                self.action_bounds,
                self.action_initial_guess,
                self.action_min,
                self.action_max,
            ) = self.handle_bounds(action_bounds, self.dim_action, 0)

        self.action = AwaitedParameter(
            "action", awaited_from=self.update_action.__name__
        )

    def __call__(self, observation: np.array) -> np.array:
        """Call the get_action method to calculate the action for the given observation.

        Args:
            observation: The observation based on which the action is to be calculated.

        Returns:
            The action calculated by the policy.
        """
        return self.get_action(observation)

    @property
    def action(self):
        """Return the action calculated by the policy."""
        return self._action

    @action.setter
    def action(self, value):
        """Get the action calculated by the policy."""
        self._action = value

    @action.getter
    def action(self):
        """Get the action calculated by the policy."""
        if isinstance(self._action, cs.DM):
            return self._action.full()
        elif isinstance(self._action, torch.Tensor):
            return self._action.detach().cpu().numpy()
        return self._action

    @property
    def data_buffer_objective_keys(self) -> Optional[List[str]]:
        """Provide the dictionary keys used for the objective function within data buffers.

        Returns:
            A list of keys used within the data buffers for objective function calculation.
        """
        pass

    @property
    def weights(self) -> Weights:
        """Get the weights of the policy model."""
        return self.model.weights

    def receive_observation(self, observation: np.array) -> None:
        """Update the current observation of the policy.

        Args:
            observation: The current observation.
        """
        self.observation = observation

    def receive_estimated_state(self, state: np.array) -> None:
        """Update the current observation of the policy.

        Args:
            state: The current state.
        """
        self.state = state

    def set_action(self, action: np.array):
        """Set the current action of the policy.

        Args:
            action: The current action.
        """
        self.action = action

    def update_action(self, observation: Optional[np.array] = None) -> np.array:
        """Update the current action of the policy based on the provided observation.

        This method uses the current model to compute a new action,
        possibly incorporating random elements for exploration purposes.
        If no observation is provided, the method uses the last received observation.

        Args:
            observation: The most recent observation received from the environment. If not
                provided, the previously received observation is used.
        """
        if observation is None:
            observation = self.observation

        if self.epsilon_random_parameter is not None:
            toss = np.random.choice(
                2,
                1,
                p=[
                    1 - self.epsilon_random_parameter,
                    self.epsilon_random_parameter,
                ],
            )
            is_exploration = bool(toss)

            if is_exploration:
                self.action = np.array(
                    [
                        np.random.uniform(self.action_min[k], self.action_max[k])
                        for k in range(len(self.action))
                    ]
                )
            else:
                self.action = self.get_action(observation)
        else:
            self.action = self.get_action(observation)

        return self.action

    def get_action(self, observation: np.array) -> np.array:
        if isinstance(self.model, ModelNN):
            action = self.model(torch.FloatTensor(observation)).detach().cpu().numpy()
        else:
            action = self.model(observation)
        return action

    def update_weights(self, weights: Optional[Weights] = None) -> None:
        """Update the weights of the model of the policy.

        Args:
            weights (numpy array, optional): The weights to update the
                model with. If not provided, the previously optimized
                weights will be used.
        """
        self.model.update_weights(weights)

    def cache_weights(self, weights: Optional[Weights] = None) -> None:
        """Cache the current weights of the model of the policy.

        Args:
            weights: The weights to cache. If not provided, the current weights will be used.
        """
        self.model.cache_weights(weights)

    def update_and_cache_weights(self, weights: Optional[Weights] = None):
        """Update and cache the weights of the model of the policy.

        Args:
            weights: The weights to update and cache. If not provided, the previously optimized weights will be used.
        """
        self.model.update_and_cache_weights(weights)

    def restore_weights(self):
        """Restore the previously cached weights of the model of thepolicy."""
        self.model.restore_weights()


class PolicyGradient(Policy, ABC):
    """Base Class for policy gradient methods.

    Base class for policy gradient methods. Policy gradients methods
    optimize policies based on the gradient of some performance measure
    with respect to policy parameters. These methods typically involve
    estimating the gradient using a large number of samples (trajectories
    for which we have either complete values or advantage estimates).
    """

    def __init__(
        self,
        model: ModelNN,
        system: Union[System, ComposedSystem],
        optimizer_config: OptimizerConfig,
        discount_factor: float = 1.0,
        device: str = "cpu",
        critic: Critic = None,
    ):
        """Instantiate PolicyGradient base class.

        Args:
            model: The neural network representing the policy.
            system: The system or environment where the policy is deployed.
            optimizer_config: Configuration for the optimization process.
            discount_factor: The discount factor for future running objectives. Defaults to 1.0, indicating no discount.
            device: The device for computations, either 'cpu' or 'cuda:0', etc..
            critic: The critic object, could be either value or action-value based.

        Raises:
            NotImplementedError: If the method is used directly without a concrete implementation in a subclass.
        """
        Policy.__init__(
            self,
            model=model,
            system=system,
            discount_factor=discount_factor,
            optimizer_config=optimizer_config,
        )
        self.device = torch.device(device)
        self.critic = critic

        self.N_episodes: int

    def update_data_buffer(self, data_buffer: DataBuffer) -> None:
        """Add precomputed values to the data buffer.

        Args:
            data_buffer: Data buffer with experience replay.
        """

        pass

    def optimize(self, data_buffer: DataBuffer) -> None:
        """optimize the policy based on the given data buffer.

        Note:
            Concrete implementations of this method should follow the optimization
            procedure set by the `optimizer_config`.

        Args:
            data_buffer: Buffer storing interaction data with the environment used for optimizing the policy.
        """

        # Send to device before optimization
        if self.critic is not None:
            self.critic.model = self.critic.model.to(self.device)
        self.model = self.model.to(self.device)
        self.N_episodes = len(np.unique(data_buffer.data["episode_id"]))
        self.update_data_buffer(data_buffer)

        super().optimize(
            **data_buffer.get_optimization_kwargs(
                keys=self.data_buffer_objective_keys(),
                optimizer_config=self.optimizer_config,
            )
        )

        # Send back to cpu after optimization
        if self.critic is not None:
            self.critic.model = self.critic.model.to(torch.device("cpu"))
        self.model = self.model.to(torch.device("cpu"))

    def initialize_optimization_procedure(self) -> None:
        """initialize all parameters necessary to carry out the optimization procedure."""
        self.objective_inputs = [
            self.create_variable(name=variable_name, is_constant=True)
            for variable_name in self.data_buffer_objective_keys()
        ]
        self.policy_weights = self.create_variable(
            name="policy_weights", like=self.model.named_parameters
        )
        self.register_objective(
            self.objective_function, variables=self.objective_inputs
        )

    @abstractmethod
    def objective_function(self, **kwargs: Any) -> torch.Tensor:
        """Abstract method defining the objective function to be optimized.

        Note:
            The objective function should be formulated based on kwargs which represent variables needed
            for computing the function such as observations, actions, estimated advantages, etc.

        Args:
            **kwargs: Additional keyword arguments that the objective function might need.

        Raises:
            NotImplementedError: If the method is used directly without concrete implementation in a subclass.
        """
        pass


class PolicyReinforce(PolicyGradient):
    """The Reinforce class extends the PolicyGradient class and implements the REINFORCE algorithm."""

    def __init__(
        self,
        model: ModelNN,
        system: Union[System, ComposedSystem],
        optimizer_config: OptimizerConfig,
        discount_factor: float = 1.0,
        device: str = "cpu",
        is_with_baseline: bool = True,
        is_do_not_let_the_past_distract_you: bool = True,
    ):
        """Instantiate Reinforce class.

        Args:
            model: The neural network-based policy model.
            system: The environment or system where the policy is deployed.
            optimizer_config: Settings for the associated optimization procedure.
            discount_factor: Discount factor for future running objectives, usually in the interval [0, 1].
            device: The device ('cpu' or 'cuda') to conduct the optimization process
            is_with_baseline (bool, optional): Indicates the use of a baseline to reduce variance in objective computation of gradients.
            is_do_not_let_the_past_distract_you: Whether to use tail values in objective.
        """
        PolicyGradient.__init__(
            self,
            model=model,
            system=system,
            optimizer_config=optimizer_config,
            discount_factor=discount_factor,
            device=device,
        )
        self.is_with_baseline = is_with_baseline
        self.is_do_not_let_the_past_distract_you = is_do_not_let_the_past_distract_you
        self.next_baseline = 0.0

        self.initialize_optimization_procedure()

    def update_data_buffer(self, data_buffer: DataBuffer) -> None:
        """Prepares the data buffer for optimization by updating values such as `"value"`, `"tail_value"`, and `"baseline"`.

        Args:
            data_buffer: The data buffer containing trajectory information for optimization.

        Note:
            Does not return any values but updates the data buffer in place.

        """
        data_buffer.update(
            {
                "value": self.calculate_last_values(data_buffer),
            }
        )
        data_buffer.update({"tail_value": self.calculate_tail_values(data_buffer)})
        data_buffer.update({"baseline": self.calculate_baseline(data_buffer)})

    def calculate_last_values(self, data_buffer: DataBuffer) -> np.array:
        """Computes the last observed value in each trajectory and maps it to all steps within the same episode.

        Args:
            data_buffer: The data buffer containing trajectory information.

        Returns:
            A numpy array containing the last value of each trajectory for all steps.
        """

        data = data_buffer.to_pandas(keys=["episode_id", "current_value"])
        data["episode_id"] = data["episode_id"].astype(int)
        data["current_value"] = data["current_value"].astype(float)
        return (
            data.groupby("episode_id")["current_value"]
            .last()
            .loc[data["episode_id"]]
            .values.reshape(-1)
        )

    def calculate_tail_values(
        self,
        data_buffer: DataBuffer,
    ) -> np.array:
        data = data_buffer.to_pandas(keys=["episode_id", "current_value"])
        data["episode_id"] = data["episode_id"].astype(int)
        data["current_value"] = data["current_value"].astype(float)

        groupby_episode_values = data.groupby(["episode_id"])["current_value"]

        last_values = (
            groupby_episode_values.last().loc[data["episode_id"]].values.reshape(-1)
        )
        current_values_shifted = groupby_episode_values.shift(
            periods=1, fill_value=0.0
        ).values.reshape(-1)

        return last_values - current_values_shifted

    def calculate_baseline(self, data_buffer: DataBuffer) -> np.array:
        r"""Computes the baseline value used for scaling the advantages during optimization.

        Args:
            data_buffer: The data buffer containing trajectory information.

        Returns:
            Array of baselines $b_{[0 : T]}$, where
                $b_t = \frac{1}{N} \sum_{j = 1} ^ {M}  \sum_{t' = t} ^ T \gamma ^ {t \delta t} r(\obs^j_t, \action^j_t)$.
                $\delta t$ is sampling time, $\obs^j_t$ and $\action^j_t$ are the observation and the action
                from step t and episode $j$ from the **previous iteration**, $\gamma$ is discount factor,
                $M$ is the number of episodes.

        Notes:
            The initial value of baseline is set to zero,
        """

        baseline = self.next_baseline
        if not self.is_do_not_let_the_past_distract_you:
            self.next_baseline = np.mean(data_buffer.to_pandas(keys=["value"]).values)

        else:
            self.next_baseline = (
                data_buffer.to_pandas(keys=["tail_value", "step_id"])
                .astype(float)
                .groupby("step_id")
                .mean()
                .loc[
                    data_buffer.to_pandas(keys=["step_id"])
                    .astype(float)
                    .values.reshape(-1)
                ]
            ).values
        if isinstance(baseline, float):
            return np.full(shape=len(data_buffer), fill_value=baseline)
        else:
            return baseline

    def data_buffer_objective_keys(self) -> List[str]:
        """Define the specific keys required from the data buffer to calculate the objective.

        Returns:
            A list of string keys corresponding to the required data: observation, action, tail_value, value, and baseline.
        """
        return [
            "observation",
            "action",
            "tail_value",
            "value",
            "baseline",
        ]

    def objective_function(
        self,
        observation: torch.Tensor,
        action: torch.Tensor,
        tail_value: torch.Tensor,
        value: torch.Tensor,
        baseline: torch.Tensor,
    ) -> torch.Tensor:
        """Calculates the surrogate loss using the reinforcement learning objective, as defined by the REINFORCE algorithm.

        Args:
            observation: The observations received from the environment.
            action: The actions taken in the environment.
            tail_value: The tail values calculated for the trajectories.
            value: The values of the current state-action pairs.
            baseline: The baseline values used for scaling.

        Returns:
            A loss value that, when minimized, leads to policy improvement.

        Note:
            All the arguments are sampled from [`DataBuffer`][regelum.data_buffers.DataBuffer]
            via [`iter_bathes`][regelum.data_buffers.DataBuffer.iter_batches] method
        """
        return reinforce_objective(
            policy_model=self.model,
            observations=observation,
            actions=action,
            tail_values=tail_value,
            values=value,
            baselines=baseline,
            is_with_baseline=self.is_with_baseline,
            is_do_not_let_the_past_distract_you=self.is_do_not_let_the_past_distract_you,
            N_episodes=self.N_episodes,
        )


class PolicySDPG(PolicyGradient):
    """Policy for Stochastic Deep Policy Gradient (SDPG)

    Extends PolicyGradient via using a critic for temporal difference objective."""

    def __init__(
        self,
        model: ModelNN,
        critic: Critic,
        system: Union[System, ComposedSystem],
        optimizer_config: OptimizerConfig,
        sampling_time: float,
        discount_factor: float = 1.0,
        device: str = "cpu",
        is_normalize_advantages: bool = True,
        gae_lambda: float = 1.0,
    ):
        """Instantiate SDPG class.

        Args:
            model: Policy model used to compute actions.
            critic: Critic object used in the temporal difference objective.
            system: The environment or system where the policy is deployed.
            optimizer_config: Settings for the associated optimization procedure.
            sampling_time: Time interval used to discretize continuous time into steps.
            discount_factor: Discount factor used to compute the present value of future running objectives.
            device: The computation device ('cpu' or 'cuda') to execute the optimization process.
            is_normalize_advantages: Whether to normalize the advantages.
            gae_lambda: Factor for trade-off of bias vs variance for Generalized Advantage Estimator
        """
        PolicyGradient.__init__(
            self,
            model=model,
            system=system,
            discount_factor=discount_factor,
            device=device,
            critic=critic,
            optimizer_config=optimizer_config,
        )
        self.sampling_time = sampling_time
        self.is_normalize_advantages = is_normalize_advantages
        self.gae_lambda = gae_lambda
        self.initialize_optimization_procedure()

    def data_buffer_objective_keys(self) -> List[str]:
        """Specifies the keys in the data buffer required for computing the optimization objective.

        Returns:
            A list of strings that are keys in the data buffer for retrieving necessary data.

        Note:
            The specified keys are 'observation', 'action', 'time', 'episode_id', and 'running_objective'.
        """
        return ["observation", "action", "time", "episode_id", "running_objective"]

    def objective_function(
        self,
        observation: torch.Tensor,
        action: torch.Tensor,
        time: torch.Tensor,
        episode_id: torch.Tensor,
        running_objective: torch.Tensor,
    ) -> torch.Tensor:
        """Defines the objective function to optimize policy under the SDPG framework using the provided data buffer components.

        Args:
            observation: The most recent observations received from the environment.
            action: The actions executed in the environment based on the policy.
            time: The timestep or moment when the action is taken.
            episode_id: Unique identifiers for episodes in the training data.
            running_objective: Cumulative reward or running objective value of the taken actions.

        Returns:
            The calculated objective value from the SDPG algorithm for optimization purposes.

        Notes:
            The objective function gets arguments from data buffer.
        """

        return sdpg_objective(
            policy_model=self.model,
            critic_model=self.critic.model,
            observations=observation,
            actions=action,
            times=time,
            discount_factor=self.discount_factor,
            N_episodes=self.N_episodes,
            episode_ids=episode_id.long(),
            running_objectives=running_objective,
            sampling_time=self.sampling_time,
            is_normalize_advantages=self.is_normalize_advantages,
            gae_lambda=self.gae_lambda,
        )


class PolicyPPO(PolicyGradient):
    """Implements the Proximal Policy Optimization (PPO) algorithm"""

    def __init__(
        self,
        model: ModelNN,
        critic: Critic,
        system: Union[System, ComposedSystem],
        optimizer_config: OptimizerConfig,
        sampling_time: float,
        discount_factor: float = 1.0,
        device: str = "cpu",
        gae_lambda: float = 0.0,
        running_objective_type: str = "cost",
        cliprange: float = 0.2,
        is_normalize_advantages: bool = True,
        entropy_coeff: float = 0.0,
    ):
        """Instantiate PPO policy class.

        Args:
            model: An instance of the policy model.
            critic: An instance of the critic used for optimization.
            system: The environment or system the policy will interact with.
            optimizer_config: Configuration parameters for the optimizer.
            sampling_time:  Sampling period used for discretizing continuous domains.
            discount_factor: Factor to discount future rewards.
            device: The device to perform computation on (CPU or GPU).
            gae_lambda: Lambda parameter for Generalized Advantage Estimation (GAE).
            running_objective_type: Type of the running objective, can be 'cost' or 'reward'.
            cliprange: Range used for clipping in PPO algorithm.
            is_normalize_advantages: Whether to normalize the advantages.
        """
        PolicyGradient.__init__(
            self,
            model=model,
            system=system,
            discount_factor=discount_factor,
            device=device,
            critic=critic,
            optimizer_config=optimizer_config,
        )
        self.sampling_time = sampling_time
        self.cliprange = cliprange
        self.running_objective_type = running_objective_type
        self.gae_lambda = gae_lambda
        self.is_normalize_advantages = is_normalize_advantages
        self.entropy_coeff = entropy_coeff
        self.initialize_optimization_procedure()

    def data_buffer_objective_keys(self) -> List[str]:
        """Return a list of data keys that will be used in the objective function of PPO optimization.

        Returns:
            A list of string keys for retrieving necessary data from the buffer for optimization.
        """

        return [
            "observation",
            "action",
            "time",
            "episode_id",
            "running_objective",
            "initial_log_probs",
        ]

    def objective_function(
        self,
        observation: torch.Tensor,
        action: torch.Tensor,
        time: torch.Tensor,
        episode_id: torch.Tensor,
        running_objective: torch.Tensor,
        initial_log_probs: torch.Tensor,
    ) -> torch.Tensor:
        """Define the surrogate objective function used for PPO optimization.

        Args:
            observation: Observations from the environment.
            action: Actions taken by the policy.
            time: Time steps of the action.
            episode_id: Identifiers for episodes within the buffer.
            running_objective: Cumulative reward or cost for each step in the episodes.
            initial_log_probs: Initial policy log probabilities of actions.

        Returns:
            The surrogate objective function value for the PPO algorithm.
        """
        return ppo_objective(
            policy_model=self.model,
            critic_model=self.critic.model,
            observations=observation,
            actions=action,
            times=time,
            discount_factor=self.discount_factor,
            N_episodes=self.N_episodes,
            episode_ids=episode_id.long(),
            running_objectives=running_objective,
            initial_log_probs=initial_log_probs,
            cliprange=self.cliprange,
            running_objective_type=self.running_objective_type,
            sampling_time=self.sampling_time,
            gae_lambda=self.gae_lambda,
            is_normalize_advantages=self.is_normalize_advantages,
            entropy_coeff=self.entropy_coeff,
        )

    def update_data_buffer(self, data_buffer: DataBuffer) -> None:
        """Perform updates to the data buffer, preparing the necessary data for PPO optimization, such as log probabilities.

        Args:
            data_buffer: The buffer containing interaction data between the agent and the environment.

        """
        data_buffer.update(
            {
                "initial_log_probs": self.model.log_pdf(
                    torch.FloatTensor(np.vstack(data_buffer.data["observation"])),
                    torch.FloatTensor(np.vstack(data_buffer.data["action"])),
                )
                .detach()
                .cpu()
                .numpy()
            }
        )


class PolicyDDPG(PolicyGradient):
    """Policy for Deterministic Deep Policy Gradient (DDPG)."""

    def __init__(
        self,
        model: ModelNN,
        critic: Critic,
        system: Union[System, ComposedSystem],
        optimizer_config: OptimizerConfig,
        discount_factor: float = 1.0,
        device: str = "cpu",
    ):
        """Instantiate DDPG class.

        Args:
            model: The neural network policy model.
            critic: The critic network used to evaluate the action value function.
            system: The environment where the policy is deployed.
            optimizer_config: Configuration for the optimization process.
            discount_factor: The discount factor for future rewards, generally between 0 and 1.
            device: The computation device, typically 'cpu' or 'cuda'.
        """
        PolicyGradient.__init__(
            self,
            model=model,
            system=system,
            discount_factor=discount_factor,
            device=device,
            critic=critic,
            optimizer_config=optimizer_config,
        )
        self.initialize_optimization_procedure()

    def data_buffer_objective_keys(self) -> List[str]:
        """Specify the keys from the data buffer that are required for computing the objective function for DDPG optimization.

        Returns:
            A list of string keys representing the required data from the data buffer.
        """
        return ["observation"]

    def objective_function(self, observation: torch.Tensor) -> torch.Tensor:
        """Calculate the objective function for DDPG, using the provided observations data to optimize the policy.

        Args:
            observation: The observations received from the environment.

        Returns:
            The objective value based on the DDPG algorithm for policy optimization.
        """

        return ddpg_objective(
            policy_model=self.model,
            critic_model=self.critic.model,
            observations=observation,
        )


class RLPolicy(Policy):
    """Class for with Predictive Control algorithms."""

    def __init__(
        self,
        model: Union[ModelNN, Model],
        critic: Critic,
        system: Union[System, ComposedSystem],
        action_bounds: Union[list, np.ndarray, None],
        optimizer_config: OptimizerConfig,
        predictor: Optional[Predictor] = None,
        prediction_horizon: Optional[int] = None,
        running_objective: Optional[RunningObjective] = None,
        constraint_parser: Optional[ConstraintParser] = None,
        discount_factor: float = 1.0,
        device: str = "cpu",
        epsilon_random_parameter: Optional[float] = None,
        algorithm: str = "mpc",
    ):
        """Initializes the RLPolicy.

        Args:
            model: The model representing the predictive policy.
            critic: The critic used for reinforcement learning, which can be value-based or action-based.
            system: The environment where the policy operates.
            action_bounds: Limits on the range of possible actions.
            optimizer_config: Configuration settings for the optimization procedure.
            predictor: Optional predictor used to forecast state sequences.
            prediction_horizon: Duration over which predictions are made.
            running_objective: The objective function used in control problems.
            constraint_parser: Optional parser for parsing and handling constraints.
            discount_factor: Discount factor for future rewards in control problems.
            device: The computational device to be used, either "cpu" or "cuda".
            epsilon_random_parameter: The probability parameter for the epsilon-greedy policy.
            algorithm: The type of RL algorithm to use, such as "mpc", "rpv", "sql", "rql", "greedy".
        """

        Policy.__init__(
            self,
            model=model,
            system=system,
            action_bounds=action_bounds,
            optimizer_config=optimizer_config,
            discount_factor=discount_factor,
        )

        self.action_bounds = np.array(action_bounds)
        self.predictor = predictor
        self.prediction_horizon = prediction_horizon
        self.critic = critic
        self.device = device
        self.epsilon_random_parameter = epsilon_random_parameter
        self.running_objective = running_objective
        self.algorithm = algorithm
        self.constraint_parser = (
            constraint_parser
            if constraint_parser is not None
            else ConstraintParserTrivial()
        )
        self.initialize_optimization_procedure()
        self.initial_guess = None

    def data_buffer_objective_keys(self) -> List[str]:
        """Keys to sample from data buffer for optimization.

        Identifies the required keys in the data buffer for calculating the optimization objective under the
        reinforcement learning policy.

        Returns:
            A list of string keys required from the data buffer for optimization.
        """
        return ["observation", "estimated_state"]

    def initialize_optimization_procedure(self):
        """Prepare the optimization environment by setting up the objective function and constraints.

        Note:
            Initializes problem-specific variables required for the optimization process, which will vary based on the algorithm used.
        """

        objective_variables = []
        self.observation_variable = self.create_variable(
            1, self.system.dim_observation, name="observation", is_constant=True
        )
        self.estimated_state_variable = self.create_variable(
            1, self.system.dim_state, name="estimated_state", is_constant=True
        )
        objective_variables.extend(
            [self.observation_variable, self.estimated_state_variable]
        )
        self.policy_model_weights = self.create_variable(
            name="policy_model_weights", like=self.model.named_parameters
        )
        if isinstance(self.model, ModelWeightContainer):
            (
                self.action_bounds_tiled,
                self.action_initial_guess,
                self.action_min,
                self.action_max,
            ) = self.handle_bounds(
                self.action_bounds,
                self.dim_action,
                tile_parameter=self.model.weights.shape[0],
            )
            self.register_bounds(self.policy_model_weights, self.action_bounds_tiled)

        objective_variables.append(self.policy_model_weights)
        if not isinstance(self.critic, CriticTrivial):
            self.critic_weights = self.create_variable(
                name="critic_weights",
                like=self.critic.model.named_parameters,
                is_constant=True,
            )
            objective_variables.append(self.critic_weights)

        self.register_objective(
            self.objective_function,
            variables=objective_variables,
        )

        if len(list(self.constraint_parser.constraint_parameters())) > 0:
            if self.algorithm == "rpv":
                is_predict_last = True
            else:
                is_predict_last = False

            self.predicted_states_var = self.create_variable(
                self.prediction_horizon,
                self.system.dim_state,
                name="predicted_states",
                is_nested_function=True,
                nested_variables=[self.policy_model_weights],
            )
            self.connect_source(
                connect_to=self.predicted_states_var,
                func=self.predictor.predict_state_sequence_from_model,
                prediction_horizon=self.prediction_horizon,
                state=self.estimated_state_variable,
                model=self.model,
                model_weights=self.policy_model_weights,
                is_predict_last=is_predict_last,
                return_predicted_states_only=True,
            )
            self.constraint_parameters = [
                self.create_variable(
                    *parameter.dims, name=parameter.name, is_constant=True
                ).with_data(parameter.data)
                for parameter in self.constraint_parser
            ]
            self.register_constraint(
                self.constraint_parser.constraint_function,
                variables=[self.predicted_states_var, *self.constraint_parameters],
            )

    def objective_function(
        self,
        estimated_state: RgArray,
        observation: RgArray,
        policy_model_weights: Weights,
        critic_weights: Optional[Weights] = None,
    ) -> RgArray:
        """
        Provides the interface for calculating the RL-based objective function for policy optimization.

        Args:
            estimated_state: The estimated current state of the system.
            observation: Latest observed state from the environment.
            policy_model_weights: Weights of the policy model, typically obtained from an optimizer.
            critic_weights: Weights of the critic model, used if the algorithm is critic-based.

        Returns:
            The calculated objective based on the specified RL algorithm.
        """
        if self.algorithm == "mpc":
            return mpc_objective(
                estimated_state=estimated_state,
                observation=observation,
                policy_model_weights=policy_model_weights,
                predictor=self.predictor,
                running_objective=self.running_objective,
                model=self.model,
                prediction_horizon=self.prediction_horizon,
                discount_factor=self.discount_factor,
            )
        elif self.algorithm == "rpv":
            return rpv_objective(
                estimated_state=estimated_state,
                observation=observation,
                policy_model_weights=policy_model_weights,
                predictor=self.predictor,
                running_objective=self.running_objective,
                model=self.model,
                prediction_horizon=self.prediction_horizon,
                discount_factor=self.discount_factor,
                critic=self.critic,
                critic_weights=critic_weights,
            )
        elif self.algorithm == "sql":
            return sql_objective(
                estimated_state=estimated_state,
                observation=observation,
                policy_model_weights=policy_model_weights,
                predictor=self.predictor,
                model=self.model,
                prediction_horizon=self.prediction_horizon,
                critic=self.critic,
                critic_weights=critic_weights,
            )
        elif self.algorithm == "rql":
            return rql_objective(
                estimated_state=estimated_state,
                observation=observation,
                policy_model_weights=policy_model_weights,
                predictor=self.predictor,
                running_objective=self.running_objective,
                model=self.model,
                prediction_horizon=self.prediction_horizon,
                critic=self.critic,
                critic_weights=critic_weights,
                discount_factor=self.discount_factor,
            )
        elif self.algorithm == "greedy":
            return rg.mean(
                self.critic(
                    observation,
                    self.model(observation, weights=policy_model_weights),
                    weights=critic_weights,
                )
            )
        else:
            raise AssertionError("RLPolicy: Wrong algorithm name")

    def optimize(self, data_buffer: DataBuffer) -> None:
        """Run the optimization procedure to update the policy model using the specified RL algorithm and data collected in the buffer.

        Args:
            data_buffer: The data buffer containing trajectories collected during policy execution.
        """
        opt_kwargs = data_buffer.get_optimization_kwargs(
            keys=self.data_buffer_objective_keys(),
            optimizer_config=self.optimizer_config,
        )

        if opt_kwargs is not None:
            if self.kind == "symbolic":
                result = (
                    super().optimize(
                        **opt_kwargs,
                        policy_model_weights=self.get_initial_guess(),
                        tol=1e-8,
                    )
                    if self.critic.weights is None
                    else super().optimize(
                        **opt_kwargs,
                        policy_model_weights=self.get_initial_guess(),
                        critic_weights=self.critic.weights,
                        tol=1e-8,
                    )
                )
                self.update_weights(result["policy_model_weights"])
            elif self.kind == "tensor":
                super().optimize(**opt_kwargs)

    def get_initial_guess(self):
        """Generates an initial guess for the policy parameters to start the optimization process.

        Returns:
            A numpy array or similar data structure representing the initial guess for policy weights.
        """

        return self.model.weights


class KinPointStabilizingPolicy(Policy):
    r"""Scenario for kinematic point stabilization.

    Implements the following policy: $\action = -\text{gain} \cdot \observation$
    """

    def __init__(self, gain):
        """Initialize an instance of the class with the given gain.

        Args:
            gain: The gain value to set for the instance.

        Returns:
            None
        """
        super().__init__()
        self.gain = gain

    def get_action(self, observation):
        return -self.gain * observation


class ThreeWheeledWRobotKinematicStabilizingPolicy(Policy):
    """Scenario for non-inertial three-wheeled robot composed of three PID scenarios."""

    def __init__(self, K):
        """Initialize an instance of scenario.

        Args:
            K: gain of scenario
        """
        super().__init__()
        self.K = K

    def get_action(self, observation):
        x = observation[0, 0]
        y = observation[0, 1]
        angle = observation[0, 2]

        angle_cond = np.arctan2(y, x)

        if not np.allclose((x, y), (0, 0), atol=1e-03) and not np.isclose(
            angle, angle_cond, atol=1e-03
        ):
            omega = (
                -self.K
                * np.sign(angle - angle_cond)
                * rg.sqrt(rg.abs(angle - angle_cond))
            )
            v = 0
        elif not np.allclose((x, y), (0, 0), atol=1e-03) and np.isclose(
            angle, angle_cond, atol=1e-03
        ):
            omega = 0
            v = -self.K * rg.sqrt(rg.norm_2(rg.hstack([x, y])))
        elif np.allclose((x, y), (0, 0), atol=1e-03) and not np.isclose(
            angle, 0, atol=1e-03
        ):
            omega = -self.K * np.sign(angle) * rg.sqrt(rg.abs(angle))
            v = 0
        else:
            omega = 0
            v = 0

        return rg.force_row(rg.hstack([v, omega]))


class InvertedPendulumStabilizingPolicy(Policy):
    """A nominal policy for inverted pendulum representing a PD controller."""

    def __init__(self, gain):
        """Initialize an instance of policy.

        Args:
            gain: gain of PID controller.
        """
        super().__init__()
        self.gain = gain

    def get_action(self, observation: np.array):
        return np.array(
            [[-((observation[0, 0]) + 0.1 * (observation[0, 1])) * self.gain]]
        )


class ThreeWheeledWRobotNIDisassembledCLFPolicy(Policy):
    """Nominal parking scenario for NI using disassembled control Lyapunov function."""

    def __init__(self, scenario_gain=10):
        """Initialize an instance of disassembled-clf scenario.

        Args:
            scenario_gain: gain of scenario
        """
        super().__init__()
        self.scenario_gain = scenario_gain

    def _zeta(self, xNI: np.array) -> np.array:
        sigma = np.sqrt(xNI[0] ** 2 + xNI[1] ** 2) + np.sqrt(abs(xNI[2]))

        nablaL = rg.zeros(3)

        nablaL[0] = (
            4 * xNI[0] ** 3
            + rg.abs(xNI[2]) ** 3
            / sigma**3
            * 1
            / np.sqrt(xNI[0] ** 2 + xNI[1] ** 2) ** 3
            * 2
            * xNI[0]
        )
        nablaL[1] = (
            4 * xNI[1] ** 3
            + rg.abs(xNI[2]) ** 3
            / sigma**3
            * 1
            / np.sqrt(xNI[0] ** 2 + xNI[1] ** 2) ** 3
            * 2
            * xNI[1]
        )
        nablaL[2] = 3 * rg.abs(xNI[2]) ** 2 * rg.sign(xNI[2]) + rg.abs(
            xNI[2]
        ) ** 3 / sigma**3 * 1 / np.sqrt(rg.abs(xNI[2])) * rg.sign(xNI[2])

        theta = 0

        sigma_tilde = (
            xNI[0] * rg.cos(theta) + xNI[1] * rg.sin(theta) + np.sqrt(rg.abs(xNI[2]))
        )

        nablaF = rg.zeros(3)

        nablaF[0] = (
            4 * xNI[0] ** 3 - 2 * rg.abs(xNI[2]) ** 3 * rg.cos(theta) / sigma_tilde**3
        )
        nablaF[1] = (
            4 * xNI[1] ** 3 - 2 * rg.abs(xNI[2]) ** 3 * rg.sin(theta) / sigma_tilde**3
        )
        nablaF[2] = (
            (
                3 * xNI[0] * rg.cos(theta)
                + 3 * xNI[1] * rg.sin(theta)
                + 2 * np.sqrt(rg.abs(xNI[2]))
            )
            * xNI[2] ** 2
            * rg.sign(xNI[2])
            / sigma_tilde**3
        )

        if xNI[0] == 0 and xNI[1] == 0:
            return nablaF
        else:
            return nablaL

    def _kappa(self, xNI: np.array) -> np.array:
        kappa_val = rg.zeros(2)

        G = rg.zeros([3, 2])
        G[:, 0] = rg.array([1, 0, xNI[1]], prototype=G)
        G[:, 1] = rg.array([0, 1, -xNI[0]], prototype=G)

        zeta_val = self._zeta(xNI)

        kappa_val[0] = -rg.abs(np.dot(zeta_val, G[:, 0])) ** (1 / 3) * rg.sign(
            rg.dot(zeta_val, G[:, 0])
        )
        kappa_val[1] = -rg.abs(np.dot(zeta_val, G[:, 1])) ** (1 / 3) * rg.sign(
            rg.dot(zeta_val, G[:, 1])
        )

        return kappa_val

    def _F(self, xNI, eta, theta):
        sigma_tilde = (
            xNI[0] * rg.cos(theta) + xNI[1] * rg.sin(theta) + np.sqrt(rg.abs(xNI[2]))
        )

        F = xNI[0] ** 4 + xNI[1] ** 4 + rg.abs(xNI[2]) ** 3 / sigma_tilde**2

        z = eta - self._kappa(xNI, theta)

        return F + 1 / 2 * rg.dot(z, z)

    def _Cart2NH(self, coords_Cart):
        xNI = rg.zeros(3)

        xc = coords_Cart[0]
        yc = coords_Cart[1]
        angle = coords_Cart[2]

        xNI[0] = angle
        xNI[1] = xc * rg.cos(angle) + yc * rg.sin(angle)
        xNI[2] = -2 * (yc * rg.cos(angle) - xc * rg.sin(angle)) - angle * (
            xc * rg.cos(angle) + yc * rg.sin(angle)
        )

        return xNI

    def _NH2ctrl_Cart(self, xNI, uNI):
        uCart = rg.zeros(2)

        uCart[0] = uNI[1] + 1 / 2 * uNI[0] * (xNI[2] + xNI[0] * xNI[1])
        uCart[1] = uNI[0]

        return uCart

    def get_action(self, observation):
        xNI = self._Cart2NH(observation[0])
        kappa_val = self._kappa(xNI)
        uNI = self.scenario_gain * kappa_val

        return self._NH2ctrl_Cart(xNI, uNI).reshape(1, -1)

    def compute_LF(self, observation):
        xNI = self._Cart2NH(observation)

        sigma = np.sqrt(xNI[0] ** 2 + xNI[1] ** 2) + np.sqrt(rg.abs(xNI[2]))
        LF_value = xNI[0] ** 4 + xNI[1] ** 4 + rg.abs(xNI[2]) ** 3 / sigma**2

        return LF_value


class MemoryPIDPolicy(Policy):
    """A base class for PID scenario.

    This scenario is able to use stored data in order to detect whether system is stabilized or not.
    """

    def __init__(
        self,
        P: float,
        I: float,
        D: float,
        setpoint: Union[np.array, float] = 0.0,
        sampling_time: float = 0.01,
        initial_point: Union[np.array, List[float]] = (-5, -5),
        buffer_length: int = 30,
    ):
        """Initialize an instance of ScenarioMemoryPID.

        Whatever

        Args:
            P: proportional gain
            I: integral gain
            D: differential gain
            setpoint: point using as target turing error evaluation
            sampling_time: time interval between two consecutive actions
            initial_point: point at which computations has begun
            buffer_length: length of stored buffer
        """
        super().__init__()
        self.P = P
        self.I = I
        self.D = D

        self.setpoint = setpoint
        self.integral = 0.0
        self.error_old = 0.0
        self.sampling_time = sampling_time
        self.initial_point = initial_point
        if isinstance(initial_point, (float, int)):
            self.observation_size = 1
        else:
            self.observation_size = len(initial_point)

        self.buffer_length = buffer_length
        self.observation_buffer = rg.ones((self.observation_size, buffer_length)) * 1e3

    def compute_error(self, process_variable):
        if isinstance(process_variable, (float, int)):
            error = process_variable - self.setpoint
        else:
            if len(process_variable) == 1:
                error = process_variable - self.setpoint
            else:
                norm = rg.norm_2(self.setpoint - process_variable)
                error = norm * rg.sign(rg.dot(self.initial_point, process_variable))
        return error

    def compute_integral(self, error):
        self.integral += error * self.sampling_time
        return self.integral

    def compute_error_derivative_numerically(self, error):
        error_derivative = (error - self.error_old) / self.sampling_time
        self.error_old = error
        return error_derivative

    def error_derivative(self, error):
        return None

    def compute_signal(
        self,
        process_variable,
        error_derivative=None,
    ):
        error = self.compute_error(process_variable)
        integral = self.compute_integral(error)

        if error_derivative is None:
            error_derivative = self.compute_error_derivative_numerically(error)

        PID_signal = -(self.P * error + self.I * integral + self.D * error_derivative)

        return PID_signal

    def set_setpoint(self, setpoint):
        self.setpoint = setpoint

    def update_observation_buffer(self, observation):
        self.observation_buffer = rg.push_vec(self.observation_buffer, observation)

    def set_initial_point(self, point):
        self.initial_point = point

    def reset(self):
        self.integral = 0.0
        self.error_old = 0.0

    def reset_buffer(self):
        self.observation_buffer = (
            rg.ones((self.observation_size, self.buffer_length)) * 1e3
        )

    def is_stabilized(self, stabilization_tollerance=1e-3):
        is_stabilized = np.allclose(
            self.observation_buffer,
            rg.rep_mat(rg.reshape(self.setpoint, (-1, 1)), 1, self.buffer_length),
            atol=stabilization_tollerance,
        )
        return is_stabilized


class ThreeWheeledRobotPIDPolicy:
    """Nominal stabilzing policy for ThreeWheeledRobotDynamic inertial system."""

    def __init__(
        self,
        state_init: np.array,
        PID_arctg_params: Tuple[float, float, float] = (10, 0.0, 3),
        PID_v_zero_params: Tuple[float, float, float] = (35, 0.0, 1.2),
        PID_x_y_origin_params: Tuple[float, float, float] = (35, 0.0, 35),
        PID_angle_origin_params: Tuple[float, float, float] = (30, 0.0, 10),
        v_to_zero_bounds: Tuple[float, float, float] = (0.0, 0.05),
        to_origin_bounds: Tuple[float, float, float] = (0.0, 0.1),
        to_arctan_bounds: Tuple[float, float, float] = (0.01, 0.2),
    ):
        """Initialize object.

        Args:
            state_init: initial state of 3wrobot_dyn
            PID_arctg_params: coefficients for PD scenario which sets
                the direction of robot to origin
            PID_v_zero_params: coefficients for PD scenario which forces
                speed to zero as robot moves to origin
            PID_x_y_origin_params: coefficients for PD scenario which
                moves robot to origin
            PID_angle_origin_params: coefficients for PD scenario which
                sets angle to zero near origin
            v_to_zero_bounds: bounds for enabling scenario which
                decelerates
            to_origin_bounds: bounds for enabling scenario which moves
                robot to origin
            to_arctan_bounds: bounds for enabling scenario which direct
                robot to origin
        """

        self.v_to_zero_bounds = v_to_zero_bounds
        self.to_origin_bounds = to_origin_bounds
        self.to_arctan_bounds = to_arctan_bounds

        self.state_init = state_init

        self.action_old = rg.zeros(2)
        self.PID_angle_arctan = MemoryPIDPolicy(
            *PID_arctg_params, initial_point=self.state_init[2]
        )
        self.PID_v_zero = MemoryPIDPolicy(
            *PID_v_zero_params, initial_point=self.state_init[3], setpoint=0.0
        )
        self.PID_x_y_origin = MemoryPIDPolicy(
            *PID_x_y_origin_params,
            setpoint=rg.array([0.0, 0.0]),
            initial_point=self.state_init[:2],
        )
        self.PID_angle_origin = MemoryPIDPolicy(
            *PID_angle_origin_params, setpoint=0.0, initial_point=self.state_init[2]
        )
        self.stabilization_tollerance = 1e-3
        self.current_F = 0
        self.current_M = 0

    def get_setpoint_for_PID_angle_arctan(self, x, y, eps=1e-8):
        if abs(x) < eps and abs(y) < eps:
            return 0.0
        return np.arctan2(y, x)

    @staticmethod
    def cdf_uniform(x, a, b):
        loc = a
        scale = 1 / (b - a)
        return np.clip(scale * (x - loc), 0, 1)

    def F_error_derivative(self, x, y, angle, v, eps=1e-8):
        if abs(x) < eps and abs(y) < eps:
            return rg.array([0.0])
        F_error_derivative = (
            v * (x * rg.cos(angle) + y * rg.sin(angle)) / rg.sqrt(x**2 + y**2)
        ) * rg.sign(rg.dot(self.PID_x_y_origin.initial_point, [x, y]))
        return rg.array([F_error_derivative])

    def get_action(self, observation):
        observation = observation[0]
        x = observation[0]
        y = observation[1]
        angle = rg.array([observation[2]])
        v = rg.array([observation[3]])
        omega = rg.array([observation[4]])
        (F_min, F_max), (M_min, M_max) = self.action_bounds[0], self.action_bounds[1]

        self.PID_x_y_origin.set_initial_point(rg.array([x, y]))
        F_error_derivative = self.F_error_derivative(x, y, angle[0], v[0])
        M_arctan_error_derivative = omega
        F_v_to_zero_error_derivative = self.current_F / self.m

        F = self.PID_x_y_origin.compute_signal(
            [x, y], error_derivative=F_error_derivative
        )
        angle_setpoint = rg.array([self.get_setpoint_for_PID_angle_arctan(x, y)])
        self.PID_angle_arctan.set_setpoint(angle_setpoint)
        M_arctan = self.PID_angle_arctan.compute_signal(
            angle, error_derivative=M_arctan_error_derivative
        )
        M_zero = self.PID_angle_origin.compute_signal(angle, error_derivative=omega[0])
        F_v_to_zero = self.PID_v_zero.compute_signal(
            v, error_derivative=F_v_to_zero_error_derivative
        )

        lbd_v_to_zero = self.cdf_uniform(
            rg.norm_2(rg.array([x, y, v[0]])),
            self.v_to_zero_bounds[0],
            self.v_to_zero_bounds[1],
        )
        lbd = self.cdf_uniform(
            rg.norm_2(rg.array([x, y])),
            self.to_origin_bounds[0],
            self.to_origin_bounds[1],
        )
        lbd_arctan = self.cdf_uniform(
            rg.abs(angle[0] - self.PID_angle_arctan.setpoint[0]),
            self.to_arctan_bounds[0],
            self.to_arctan_bounds[1],
        )
        control_to_origin = rg.array(
            [
                (1 - lbd_arctan) * np.clip(F[0], F_min, F_max)
                + lbd_arctan * np.clip(F_v_to_zero[0], F_min, F_max),
                lbd_arctan * np.clip(M_arctan[0], M_min, M_max),
            ]
        )
        control_v_to_zero = rg.array([np.clip(F_v_to_zero[0], F_min, F_max), 0.0])
        control_angle_to_zero = rg.array([0, np.clip(M_zero[0], M_min, M_max)])

        action = (
            lbd * control_to_origin
            + (1 - lbd) * control_v_to_zero
            + control_angle_to_zero * (1 - lbd_v_to_zero)
        )

        self.current_F = action[0]
        self.current_M = action[1]

        return action.reshape(1, -1)


class CartPoleEnergyBasedPolicy(Policy):
    """An energy-based scenario for cartpole."""

    def __init__(
        self,
        scenario_gain: float = 10,
        upright_gain=None,
        swingup_gain=10,
        pid_loc_thr=0.35,
        pid_scale_thr=10.0,
        clip_bounds=(-1, 1),
    ):
        """Initialize an instance of ScenarioCartPoleEnergyBased.

        Args:
            action_bounds: upper and lower bounds for action yielded
                from policy
            sampling_time: time interval between two consecutive actions
            scenario_gain: scenario gain
            system: an instance of Cartpole system
        """
        super().__init__()
        from regelum.system import CartPole

        self.scenario_gain = scenario_gain
        self.m_c, self.m_p, self.g, self.l = (
            CartPole().parameters["m_c"],
            CartPole().parameters["m_p"],
            CartPole().parameters["g"],
            CartPole().parameters["l"],
        )
        self.upright_gain = upright_gain
        self.swingup_gain = swingup_gain
        self.pid_loc_thr = pid_loc_thr
        self.pid_scale_thr = pid_scale_thr
        self.clip_bounds = clip_bounds

    def get_action(self, observation):
        observation = observation[0]

        # sin_theta, one_minus_cos_theta, x, theta_dot, x_dot = observation
        sin_theta, one_minus_cos_theta, theta_dot, x_dot = observation
        x = 0
        cos_theta = 1 - one_minus_cos_theta
        theta = rg.atan2(sin_theta, cos_theta)
        # E_total = (
        #     self.m_p * self.l**2 * theta_dot**2 / 2
        #     + self.m_p * self.g * self.l * (1 - cos_theta)
        # )
        # print(E_total)
        lbd = (1 - rg.tanh((theta - self.pid_loc_thr) * self.pid_scale_thr)) / 2

        low, high = self.clip_bounds
        x_clipped = rg.clip(x, low, high)
        x_dot_clipped = rg.clip(x_dot, low, high)
        # self.action = (1 - lbd) * (
        #     self.swingup_gain * E_total * rg.sign(cos_theta * theta_dot)
        # ) + lbd * self.upright_gain.T @ rg.array(
        #     [theta, x_clipped, theta_dot, x_dot_clipped]
        # )

        # self.action = self.upright_gain.T @ rg.array(
        #     [theta, x_clipped, theta_dot, x_dot_clipped]
        # )
        self.upswing_gain = 3
        if cos_theta < 0:
            action_upswing = rg.sign(theta_dot) * self.upswing_gain
        else:
            action_upswing = rg.sign(sin_theta) * self.upswing_gain

        # if (cos_theta > 0 and theta_dot > 0) or (cos_theta > 0 and theta_dot < 0):
        #     action_upswing = -1 * action_upswing

        # self.upright_gain = np.array([50, 2., 20., 3.])
        # self.upright_gain = np.array([500, 0.0, 200.0, 0.0])

        action_upright = self.upright_gain.T @ rg.array(
            [theta, x_clipped, theta_dot, x_dot_clipped]
        )

        self.action = (1 - lbd) * action_upswing + lbd * action_upright

        return self.action.reshape(1, -1)


class ConstantPolicy(Policy):
    def __init__(self, value):
        super().__init__()
        self.value = np.array(value).reshape(1, -1)

    def get_action(self, observation):
        return self.value


class LunarLanderPIDPolicy(Policy):
    """Nominal PID scenario for lunar lander."""

    def __init__(
        self,
        state_init,
        PID_angle_parameters=None,
        PID_height_parameters=None,
        PID_x_parameters=None,
    ):
        """Initialize an instance of PID scenario for lunar lander.

        Args:
            action_bounds: upper and lower bounds for action yielded
                from policy
            state_init: state at which simulation has begun
            sampling_time: time interval between two consecutive actions
            PID_angle_parameters: parameters for PID scenario
                stabilizing angle of lander
            PID_height_parameters: parameters for PID scenario
                stabilizing y-coordinate of lander
            PID_x_parameters: parameters for PID scenario stabilizing
                x-coordinate of lander
        """
        super().__init__()

        if PID_angle_parameters is None:
            PID_angle_parameters = [1, 0, 0]
        if PID_height_parameters is None:
            PID_height_parameters = [10, 0, 0]
        if PID_x_parameters is None:
            PID_x_parameters = [10, 0, 0]
        self.PID_angle = MemoryPIDPolicy(
            *PID_angle_parameters,
            initial_point=rg.array([state_init[2]]),
            setpoint=rg.array([0]),
        )
        self.PID_height = MemoryPIDPolicy(
            *PID_height_parameters,
            initial_point=rg.array([state_init[1]]),
            setpoint=rg.array([0]),
        )
        self.PID_x = MemoryPIDPolicy(
            *PID_x_parameters,
            initial_point=rg.array([state_init[2]]),
            setpoint=rg.array([0]),
        )

    def get_action(self, observation):
        observation = observation[0]
        self.action = [0, 0]

        if abs(observation[2]) > self.threshold:
            self.threshold = self.threshold_1
            self.action[0] = self.PID_angle.compute_signal(
                rg.array([observation[2]]), error_derivative=observation[5]
            )[0]

        else:
            self.threshold = self.threshold_2
            self.action[0] = self.PID_x.compute_signal(
                rg.array([observation[0]]), error_derivative=observation[3]
            )[0]
            self.action[1] = self.PID_height.compute_signal(
                rg.array([observation[1]]), error_derivative=observation[4]
            )[0]

        self.action = rg.array(self.action).reshape(1, -1)

        return self.action


class LunarLanderPIDPolicyNew(Policy):
    """Nominal PID scenario for lunar lander."""

    def __init__(
        self,
        state_init,
        PID_angle_parameters=None,
        PID_omega_parameters=None,
        PID_height_parameters=None,
        PID_y_dot_parameters=None,
        PID_x_parameters=None,
    ):
        """Initialize an instance of PID scenario for lunar lander.

        Args:
            action_bounds: upper and lower bounds for action yielded
                from policy
            state_init: state at which simulation has begun
            sampling_time: time interval between two consecutive actions
            PID_angle_parameters: parameters for PID scenario
                stabilizing angle of lander
            PID_height_parameters: parameters for PID scenario
                stabilizing y-coordinate of lander
            PID_x_parameters: parameters for PID scenario stabilizing
                x-coordinate of lander
        """
        super().__init__()
        self.PID_omega = MemoryPIDPolicy(
            *PID_omega_parameters,
            initial_point=rg.array([state_init[5]]),
            setpoint=rg.array([0]),
        )
        self.PID_angle = MemoryPIDPolicy(
            *PID_angle_parameters,
            initial_point=rg.array([state_init[5]]),
            setpoint=rg.array([0]),
        )
        self.PID_y_dot = MemoryPIDPolicy(
            *PID_y_dot_parameters,
            initial_point=rg.array([state_init[4]]),
            setpoint=rg.array([0]),
        )
        self.PID_height = MemoryPIDPolicy(
            *PID_height_parameters,
            initial_point=rg.array([state_init[1]]),
            setpoint=rg.array([5]),
        )
        self.PID_x = MemoryPIDPolicy(
            *PID_x_parameters,
            initial_point=rg.array([state_init[0]]),
            setpoint=rg.array([0]),
        )
        self.action = rg.array([0.0, 0.0]).reshape(1, 2)

    def get_action(self, observation):
        new_support_lvl = rg.min(
            rg.array(
                [
                    10
                    * rg.sqrt(
                        rg.norm_2(
                            rg.hstack((observation[0, 2:5], observation[0, 0])) ** 2
                        )
                    ),
                    3.0,
                ]
            )
        )
        self.PID_height.set_setpoint(new_support_lvl)

        self.action[0, 0] = self.PID_omega.compute_signal(
            rg.array([observation[0, 2]]),
            error_derivative=rg.array([observation[0, 5]]),
        ).reshape(-1) - rg.cos(observation[0, 2]) ** 2 * self.PID_x.compute_signal(
            rg.array([observation[0, 0]]),
            error_derivative=rg.array([observation[0, 3]]),
        ).reshape(
            -1
        )
        self.action[0, 1] = rg.sign(
            rg.cos(observation[0, 2])
        ) * self.PID_y_dot.compute_signal(rg.array([observation[0, 4]])).reshape(
            -1
        ) + rg.sign(
            rg.cos(observation[0, 2])
        ) * self.PID_height.compute_signal(
            rg.array([observation[0, 1]]),
            error_derivative=rg.array([observation[0, 4]]),
        ).reshape(
            -1
        )

        return self.action


class TwoTankPIDPolicy(Policy):
    """PID scenario for double tank system."""

    def __init__(
        self,
        state_init: np.array = None,
        PID_2tank_parameters_x1=(1, 0, 0),
        PID_2tank_parameters_x2=(1, 0, 0),
    ):
        """Initialize an instance of Scenario2TankPID.

        Args:
            action_bounds: upper and lower bounds for action yielded
                from policy
            params: parameters of double tank system
            state_init: state at which simulation has begun
            sampling_time: time interval between two consecutive actions
            PID_2tank_parameters_x1: parameters for PID scenario
                stabilizing first component of system's state
            PID_2tank_parameters_x2: parameters for PID scenario
                stabilizing second component of system's state
        """
        from regelum.system import TwoTank

        super().__init__()
        params = TwoTank._parameters

        self.K1 = params["K1"]
        self.K2 = params["K2"]
        self.K3 = params["K3"]
        self.tau1 = params["tau1"]
        self.tau2 = params["tau2"]

        self.state_init = state_init
        self.PID_2tank_x1 = MemoryPIDPolicy(
            *PID_2tank_parameters_x1,
            initial_point=rg.array([state_init[0]]),
        )
        self.PID_2tank_x2 = MemoryPIDPolicy(
            *PID_2tank_parameters_x2,
            initial_point=rg.array([state_init[1]]),
        )

        self.action = np.zeros((1, 1))

    def get_action(self, observation):
        error_derivative_x1 = -(
            1 / (self.tau1) * (-observation[0][0] + self.K1 * self.action[0][0])
        )
        error_derivative_x2 = (
            -1
            / (self.tau2)
            * (
                -observation[0][1]
                + self.K2 * observation[0][0]
                + self.K3 * observation[0][1] ** 2
            )
        )

        self.action = self.PID_2tank_x1.compute_signal(
            rg.array([[observation[0][0]]]), error_derivative=error_derivative_x1
        ) + self.PID_2tank_x2.compute_signal(
            rg.array([[observation[0][1]]]), error_derivative=error_derivative_x2
        )
        return self.action


class ThreeWheeledRobotDisassembledCLFPolicy(Policy):
    """Nominal scenario for 3-wheel robots used for benchmarking of other scenarios.

    The scenario is sampled.

    For a 3-wheel robot with dynamical pushing force and steering torque (a.k.a. ENDI - extended non-holonomic double integrator) [[1]_], we use here
    a scenario designed by non-smooth backstepping (read more in [[2]_], [[3]_]).

    Attributes:
    m, moment_of_inertia : : numbers
        Mass and moment of inertia around vertical axis of the robot.
    controller_gain : : number
        Controller gain.
        Initial value of the controller's __internal clock.
    sampling_time : : number
        Scenario's sampling time (in seconds).

    References:
    .. [1] W. Abbasi, F. urRehman, and I. Shah. “Backstepping based nonlinear adaptive control for the extended
           nonholonomic double integrator”. In: Kybernetika 53.4 (2017), pp. 578–594

    .. [2] Matsumoto, R., Nakamura, H., Satoh, Y., and Kimura, S. (2015). Position control of two-wheeled mobile robot
             via semiconcave function backstepping. In 2015 IEEE Conference on Control Applications (CCA), 882–887

    .. [3] Osinenko, Pavel, Patrick Schmidt, and Stefan Streif. "Nonsmooth stabilization and its computational asetpointects." arXiv preprint arXiv:2006.14013 (2020)

    """

    def __init__(
        self,
        optimizer_config,
        scenario_gain=10,
    ):
        """Initialize an instance of stabilizing policy for three wheeled robot."""
        super().__init__(optimizer_config=optimizer_config)
        from regelum.system import ThreeWheeledRobotDynamic

        self.m = ThreeWheeledRobotDynamic.parameters["m"]
        self.moment_of_inertia = ThreeWheeledRobotDynamic.parameters["I"]
        self.scenario_gain = scenario_gain
        self.xNI_var = self.create_variable(3, 1, name="xNI", is_constant=True)
        self.eta_var = self.create_variable(2, 1, name="eta", is_constant=True)
        self.theta_var = self.create_variable(
            1, 1, name="theta", is_constant=False, like=np.array([0])
        )
        self.register_bounds(self.theta_var, rg.array([[-np.pi, np.pi]]))
        self.register_objective(
            self._Fc, variables=[self.xNI_var, self.eta_var, self.theta_var]
        )

    def _zeta(self, xNI, theta):
        """Compute generic, i.e., theta-dependent, supper_bound_constraintradient (disassembled) of a CLF for NI (a.k.a. nonholonomic integrator, a 3wheel robot with static actuators)."""
        sigma_tilde = (
            xNI[0] * rg.cos(theta) + xNI[1] * rg.sin(theta) + np.sqrt(rg.abs(xNI[2]))
        )

        nablaF = rg.zeros(3, prototype=theta)

        nablaF[0] = (
            4 * xNI[0] ** 3 - 2 * rg.abs(xNI[2]) ** 3 * rg.cos(theta) / sigma_tilde**3
        )

        nablaF[1] = (
            4 * xNI[1] ** 3 - 2 * rg.abs(xNI[2]) ** 3 * rg.sin(theta) / sigma_tilde**3
        )

        nablaF[2] = (
            (
                3 * xNI[0] * rg.cos(theta)
                + 3 * xNI[1] * rg.sin(theta)
                + 2 * rg.sqrt(rg.abs(xNI[2]))
            )
            * xNI[2] ** 2
            * rg.sign(xNI[2])
            / sigma_tilde**3
        )

        return nablaF

    def _kappa(self, xNI, theta):
        """Stabilizing scenario for NI-part."""
        G = rg.zeros([2, 3], prototype=xNI)
        G[0, :] = rg.hstack([1, 0, xNI[1]])
        G[1, :] = rg.hstack([0, 1, -xNI[0]])
        G = G.T

        kappa_val = rg.zeros(2, prototype=xNI)

        zeta_val = self._zeta(xNI, theta)

        kappa_val[0] = -rg.abs(rg.dot(zeta_val, G[:, 0])) ** (1 / 3) * rg.sign(
            rg.dot(zeta_val, G[:, 0])
        )
        kappa_val[1] = -rg.abs(rg.dot(zeta_val, G[:, 1])) ** (1 / 3) * rg.sign(
            rg.dot(zeta_val, G[:, 1])
        )

        return kappa_val

    def _Fc(self, xNI, eta, theta):
        """Marginal function for ENDI constructed by nonsmooth backstepping. See details in the literature mentioned in the class documentation."""
        sigma_tilde = (
            xNI[0] * rg.cos(theta) + xNI[1] * rg.sin(theta) + rg.sqrt(rg.abs(xNI[2]))
        )

        F = xNI[0] ** 4 + xNI[1] ** 4 + rg.abs(xNI[2]) ** 3 / sigma_tilde**2

        z = eta - self._kappa(xNI, theta)

        return F + 1 / 2 * rg.dot(z, z)

    def _Cart2NH(self, coords_Cart):
        r"""Transform from Cartesian coordinates to non-holonomic (NH) coordinates.

        See Section VIII.A in [[1]_].

        The transformation is a bit different since the 3rd NI eqn reads for our case as: :math:`\dot x_3 = x_2 u_1 - x_1 u_2`.

        References:
        .. [1] Watanabe, K., Yamamoto, T., Izumi, K., & Maeyama, S. (2010, October). Underactuated control for nonholonomic mobile robots by using double
               integrator model and invariant manifold theory. In 2010 IEEE/RSJ International Conference on Intelligent Robots and Systems (pp. 2862-2867)

        """
        xNI = rg.zeros(3)
        eta = rg.zeros(2)

        xc = coords_Cart[0]
        yc = coords_Cart[1]
        angle = coords_Cart[2]
        v = coords_Cart[3]
        omega = coords_Cart[4]

        xNI[0] = angle
        xNI[1] = xc * rg.cos(angle) + yc * rg.sin(angle)
        xNI[2] = -2 * (yc * rg.cos(angle) - xc * rg.sin(angle)) - angle * (
            xc * rg.cos(angle) + yc * rg.sin(angle)
        )

        eta[0] = omega
        eta[1] = (yc * rg.cos(angle) - xc * rg.sin(angle)) * omega + v

        return [xNI, eta]

    def _NH2ctrl_Cart(self, xNI, eta, uNI):
        r"""Get control for Cartesian NI from NH coordinates.

        See Section VIII.A in [[1]_].

        The transformation is a bit different since the 3rd NI eqn reads for our case as: :math:`\dot x_3 = x_2 u_1 - x_1 u_2`.

        References:
        .. [1] Watanabe, K., Yamamoto, T., Izumi, K., & Maeyama, S. (2010, October). Underactuated control for nonholonomic mobile robots by using double
               integrator model and invariant manifold theory. In 2010 IEEE/RSJ International Conference on Intelligent Robots and Systems (pp. 2862-2867)

        """
        uCart = rg.zeros(2)

        uCart[0] = self.m * (
            uNI[1]
            + xNI[1] * eta[0] ** 2
            + 1 / 2 * (xNI[0] * xNI[1] * uNI[0] + uNI[0] * xNI[2])
        )
        uCart[1] = self.moment_of_inertia * uNI[0]

        return uCart

    def get_action(self, observation):
        """Perform the same computation as :func:`~Scenario3WRobotDisassembledCLF.compute_action`, but without invoking the __internal clock."""
        observation = observation[0]
        xNI, eta = self._Cart2NH(observation)
        theta_star = self.optimize(xNI=xNI, eta=eta)
        if self.kind == "symbolic":
            theta_star = theta_star["theta"]
        kappa_val = self._kappa(xNI, theta_star)
        z = eta - kappa_val
        uNI = -self.scenario_gain * z
        action = self._NH2ctrl_Cart(xNI, eta, uNI)
        self.action_old = action
        return action.reshape(1, -1)

    def compute_LF(self, observation):
        xNI, eta = self._Cart2NH(observation)
        theta_star = self._minimizer_theta(xNI, eta)
        return self._Fc(xNI, eta, theta_star)



"""Contains auxiliary tools."""

import inspect
import numpy as np


from enum import IntEnum
from typing import Union
from dataclasses import dataclass
from typing import Optional
import random

try:
    import casadi

    CASADI_TYPES = tuple(
        x[1] for x in inspect.getmembers(casadi.casadi, inspect.isclass)
    )
except (ModuleNotFoundError, AttributeError):
    CASADI_TYPES = tuple()
import types

try:
    import torch

    TORCH_TYPES = tuple(
        x[1]
        for x in inspect.getmembers(torch, inspect.isclass)
        if ("torch" in str(x[1]))
    )
except ModuleNotFoundError:
    TORCH_TYPES = tuple()


# TODO: MISSING ENTRANCE SENTENCE: this class is bla-bla, it is needed to do bla-bla
class RCType(IntEnum):
    """Type inference proceeds by priority: `Torch` type has priority 3, `CasADi` type has priority 2, `NumPy` type has priority 1.

    That is, if, for instance, a function of two arguments gets an argument of a `NumPy` type and an argument of a `CasAdi` type,
    then the function's output type is inferred as a `CasADi` type.
    Mixture of CasADi types will raise a `TypeError` exception.
    """

    TORCH = 3
    CASADI = 2
    NUMPY = 1


TORCH = RCType.TORCH
CASADI = RCType.CASADI
NUMPY = RCType.NUMPY


def torch_safe_log(x, eps=1e-10):
    return torch.log(x + eps)


def is_CasADi_typecheck(*args) -> Union[RCType, bool]:
    return CASADI if any([isinstance(arg, CASADI_TYPES) for arg in args]) else False


def is_Torch_typecheck(*args) -> Union[RCType, bool]:
    return TORCH if any([isinstance(arg, TORCH_TYPES) for arg in args]) else False


# TODO: ADD DOCSTRING
def type_inference(*args, **kwargs) -> Union[RCType, bool]:
    is_CasADi = is_CasADi_typecheck(*args, *kwargs.values())
    is_Torch = is_Torch_typecheck(*args, *kwargs.values())
    if is_CasADi + is_Torch > 4:
        raise TypeError(
            "There is no support for simultaneous usage of both NumPy and CasADi"
        )
    else:
        result_type = max(is_CasADi, is_Torch, NUMPY)
        return result_type


# TODO: ADD DOCSTRING
def safe_unpack(argin):
    if isinstance(argin, (list, tuple)):
        return argin
    else:
        return (argin,)


# TODO: ADD DOCSTRING
def decorateAll(decorator):
    class MetaClassDecorator(type):
        def __new__(cls, classname, supers, classdict):
            for name, elem in classdict.items():
                if (
                    isinstance(elem, types.FunctionType)
                    and (name != "__init__")
                    and not isinstance(elem, staticmethod)
                ):
                    classdict[name] = decorator(classdict[name])
            return type.__new__(cls, classname, supers, classdict)

    return MetaClassDecorator


# TODO: ADD DOCSTRING
@decorateAll
def metaclassTypeInferenceDecorator(function):
    def wrapper(*args, **kwargs):
        rc_type = kwargs.get("rc_type")
        if rc_type is not None:
            del kwargs["rc_type"]
            return function(*args, **kwargs, rc_type=rc_type)
        else:
            return function(*args, **kwargs, rc_type=type_inference(*args, **kwargs))

    return wrapper


class Clock:
    def __init__(self, period: float, eps=1e-7):
        self.period = period
        self.eps = eps
        self.reset()

    def check_time(self, time: float):
        self.delta_time = time - self.current_time
        self.current_time = time

        if (
            self.is_first_time_called
            or self.current_time > self.last_sampled_time + self.period - self.eps
        ):
            self.last_sampled_time = time
            result = True
        else:
            result = False

        self.is_first_time_called = False
        return result

    def reset(self):
        self.last_sampled_time = self.current_time = 0.0
        self.is_first_time_called = True
        self.delta_time = 0.0


@dataclass
class AwaitedParameter:
    """A nested class to tag parameters that are expected to be computed at initial simulation step."""

    name: str
    awaited_from: Optional[str]
    _string_to_show: str = ""

    def __post_init__(self):
        self._string_to_show = f"Parameter {self.name}'s value is awaited from the source {self.awaited_from}"

    def __repr__(self):
        return self._string_to_show

    def __str__(self):
        return self._string_to_show

    def __getattribute__(self, name):
        if name in [
            "__repr__",
            "__str__",
            "name",
            "awaited_from",
            "_string_to_show",
            "__post_init__",
            "__instancecheck__",
            "__reduce_ex__",
            "__reduce__",
            "__getstate__",
            "__setstate__",
            "__dict__",
            "__class__",
            "__deepcopy__",
        ]:
            return object.__getattribute__(self, name)
        else:
            raise Exception(str(self) + f" Attempted to get {name}")


# TODO: ADD DOCSTRING
class RCTypeHandler(metaclass=metaclassTypeInferenceDecorator):
    TORCH = RCType.TORCH
    CASADI = RCType.CASADI
    NUMPY = RCType.NUMPY

    def LeakyReLU(self, x, negative_slope=0.01, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.maximum(0, x) + negative_slope * np.minimum(0, x)
        elif rc_type == TORCH:
            return torch.nn.LeakyReLU(negative_slope=negative_slope)(x)
        elif rc_type == CASADI:
            return self.max(
                [self.zeros(self.shape(x), prototype=x), x]
            ) + negative_slope * self.min([self.zeros(self.shape(x), prototype=x), x])

    def CasADi_primitive(self, type: str = "MX", rc_type: RCType = NUMPY):
        if type == "MX":
            return casadi.MX.sym("x", 1)
        elif type == "SX":
            return casadi.SX.sym("x", 1)
        elif type == "DM":
            return casadi.DM([0])

    def cos(self, x, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.cos(x)
        elif rc_type == TORCH:
            return torch.cos(x)
        elif rc_type == CASADI:
            return casadi.cos(x)

    def atan2(self, x, y, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.arctan2(x, y)
        elif rc_type == TORCH:
            return torch.atan2(x, y)
        elif rc_type == CASADI:
            return casadi.atan2(x, y)

    def clip(self, x, l, u, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.clip(x, l, u)
        elif rc_type == TORCH:
            return torch.clip(x, l, u)
        elif rc_type == CASADI:
            return casadi.fmin(
                self.concatenate([casadi.max(self.concatenate([x, l])), u])
            )

    def diag(self, x, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            diag = np.diag(x)
            if len(diag.shape) == 1:
                return diag.reshape(-1, 1)
            else:
                return diag
        elif rc_type == TORCH:
            diag = torch.diag(x)
            if len(diag.shape) == 1:
                return diag.reshape(-1, 1)
            else:
                return diag
        elif rc_type == CASADI:
            return casadi.diag(x)

    def sin(self, x, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.sin(x)
        elif rc_type == TORCH:
            return torch.sin(x)
        elif rc_type == CASADI:
            return casadi.sin(x)

    def floor(self, x, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.floor(x)
        elif rc_type == TORCH:
            return torch.floor(x)
        elif rc_type == CASADI:
            return casadi.floor(x)

    def column_stack(self, tup, rc_type: RCType = NUMPY):
        rc_type = type_inference(*tup)

        if rc_type == NUMPY:
            return np.column_stack(tup)
        elif rc_type == TORCH:
            return torch.column_stack(tup)
        elif rc_type == CASADI:
            return casadi.horzcat(*tup)

    def hstack(self, tup, rc_type: RCType = NUMPY):
        rc_type = type_inference(*tup)

        if rc_type == NUMPY:
            return np.hstack(tup)
        elif rc_type == TORCH:
            return torch.hstack(tup)
        elif rc_type == CASADI:
            return casadi.horzcat(*tup)

    def vstack(self, tup, rc_type: RCType = NUMPY):
        rc_type = type_inference(*tup)

        if rc_type == NUMPY:
            return np.vstack(tup)
        elif rc_type == TORCH:
            return torch.vstack(tup)
        elif rc_type == CASADI:
            return casadi.vertcat(*tup)

    def exp(self, x, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.exp(x)
        elif rc_type == TORCH:
            return torch.exp(x)
        elif rc_type == CASADI:
            return casadi.exp(x)

    def log(self, x, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.log(x)
        elif rc_type == TORCH:
            return torch.log(x)
        elif rc_type == CASADI:
            return casadi.log(x)

    def penalty_function(
        self, x, penalty_coeff=1, delta=1, lvl=1, rc_type: RCType = NUMPY
    ):
        return self.exp(penalty_coeff * (x - delta)) - lvl

    def push_vec(self, matrix, vec, rc_type: RCType = NUMPY):
        return self.column_stack([matrix[:, 1:], vec], rc_type=rc_type)

    def reshape_CasADi_as_np(self, array, dim_params, rc_type: RCType = NUMPY):
        result = self.zeros(dim_params, prototype=array)
        n_rows, n_cols = dim_params
        array_n_rows, array_n_cols = self.shape(array)

        for i in range(n_rows):
            for j in range(n_cols):
                result[i, j] = array[
                    (i * n_cols + j) // array_n_cols, (i * n_cols + j) % array_n_cols
                ]

        return result

    def reshape_to_column(self, array, length, rc_type: RCType = NUMPY):
        result_array = rg.reshape(array, [length, 1])
        return result_array

    def reshape(
        self, array, dim_params: Union[list, tuple, int], rc_type: RCType = NUMPY
    ):
        if rc_type == CASADI:
            if isinstance(dim_params, (list, tuple)):
                if len(dim_params) > 1:
                    return self.reshape_CasADi_as_np(array, dim_params)
                else:
                    return casadi.reshape(array, dim_params[0], 1)
            elif isinstance(dim_params, int):
                return casadi.reshape(array, dim_params, 1)
            else:
                raise TypeError(
                    "Wrong type of dimension parameter was passed.\
                         Possible cases are: int, [int], [int, int, ...]"
                )

        elif rc_type == NUMPY:
            return np.reshape(array, dim_params)

        elif rc_type == TORCH:
            return torch.reshape(array, dim_params)

    def array(
        self, array, prototype=None, rc_type: RCType = NUMPY, _force_numeric=False
    ):
        if isinstance(prototype, (list, tuple)):
            rc_type = type_inference(*prototype)

        if rc_type == NUMPY:
            self._array = np.array(array)
        elif rc_type == TORCH:
            if hasattr(prototype, "device"):
                device = prototype.device
            elif hasattr(array, "device"):
                device = array.device
            else:
                device = torch.device("cpu")

            self._array = torch.FloatTensor(array).to(device=device)
        elif rc_type == CASADI:
            if _force_numeric:
                self._array = casadi.DM(array)
            else:
                casadi_constructor = (
                    type(prototype) if prototype is not None else casadi.DM
                )

                self._array = casadi_constructor(array)

        return self._array

    def ones(
        self,
        argin,
        prototype=None,
        rc_type: RCType = NUMPY,
    ):
        if isinstance(prototype, (list, tuple)):
            rc_type = type_inference(*prototype)

        if rc_type == NUMPY:
            self._array = np.ones(argin)
        elif rc_type == TORCH:
            self._array = torch.ones(argin)
        elif rc_type == CASADI:
            if isinstance(prototype, (list, tuple)):
                casadi_constructor = casadi.DM

                for constructor_type in map(lambda x: type(x), prototype):
                    if constructor_type == casadi.MX:
                        casadi_constructor = casadi.MX
                        break
            else:
                casadi_constructor = (
                    type(prototype) if prototype is not None else casadi.DM
                )

            self._array = casadi_constructor.ones(*safe_unpack(argin))

        return self._array

    def zeros(
        self,
        argin,
        prototype=None,
        rc_type: RCType = NUMPY,
    ):
        if isinstance(prototype, (list, tuple)):
            rc_type = type_inference(*prototype)

        if rc_type == NUMPY:
            return np.zeros(argin)
        elif rc_type == TORCH:
            return torch.zeros(argin)
        elif rc_type == CASADI:
            if isinstance(prototype, (list, tuple)):
                casadi_constructor = casadi.DM

                for constructor_type in map(lambda x: type(x), prototype):
                    if constructor_type == casadi.MX:
                        casadi_constructor = casadi.MX
                        break
            else:
                casadi_constructor = (
                    type(prototype) if prototype is not None else casadi.DM
                )

            self._array = casadi_constructor.zeros(*safe_unpack(argin))

            return self._array

    def concatenate(self, argin, rc_type: Union[RCType, bool] = NUMPY, axis=0):
        rc_type = type_inference(*safe_unpack(argin))
        if rc_type == NUMPY:
            return np.concatenate(argin, axis=axis)
        elif rc_type == TORCH:
            return torch.cat(argin, dim=axis)
        elif rc_type == CASADI:
            if isinstance(argin, (list, tuple)):
                if axis == 0:
                    return casadi.vertcat(*argin)
                elif axis == 1:
                    return casadi.horzcat(*argin)
                else:
                    raise ValueError("Not implemented value of axis for CasADi")

    def atleast_1d(self, dim, rc_type: RCType = NUMPY):
        return np.atleast_1d(dim)

    def transpose(self, A, rc_type: RCType = NUMPY):
        if rc_type == TORCH:
            return A.mT if len(A.shape) > 1 else A.T
        else:
            return A.T

    def vec(self, expr, rc_type: RCType = NUMPY):
        if rc_type == CASADI:
            return casadi.vec(expr)
        else:
            return expr

    def rep_mat(self, array, n, m, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.tile(array, (n, m))
        elif rc_type == TORCH:
            return torch.tile(array, (n, m))
        elif rc_type == CASADI:
            return casadi.repmat(array, n, m)

    def matmul(self, A, B, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.matmul(A, B)
        elif rc_type == TORCH:
            A = torch.tensor(A).double()
            B = torch.tensor(B).double()
            return torch.matmul(A, B)
        elif rc_type == CASADI:
            return casadi.mtimes(A, B)

    def casadi_outer(self, v1, v2, rc_type: RCType = NUMPY):
        if not is_CasADi_typecheck(v1):
            v1 = self.array_symb(v1)

        return casadi.horzcat(*[v1 * v2_i for v2_i in v2.nz])

    def outer(self, v1, v2, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.outer(v1, v2)
        elif rc_type == TORCH:
            return torch.outer(v1, v2)
        elif rc_type == CASADI:
            return self.casadi_outer(v1, v2)

    def sign(self, x, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.sign(x)
        elif rc_type == TORCH:
            return torch.sign(x)
        elif rc_type == CASADI:
            return casadi.sign(x)

    def abs(self, x, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.abs(x)
        elif rc_type == TORCH:
            return torch.abs(x)
        elif rc_type == CASADI:
            return casadi.fabs(x)

    def min(self, array, rc_type: RCType = NUMPY):
        if isinstance(array, (list, tuple)):
            rc_type = type_inference(*array)

        if rc_type == NUMPY:
            return np.min(array)
        elif rc_type == TORCH:
            return torch.min(array)
        elif rc_type == CASADI:
            return casadi.mmin(*safe_unpack(array))

    def max(self, array, rc_type: RCType = NUMPY):
        if isinstance(array, (list, tuple)):
            rc_type = type_inference(*array)

        if rc_type == NUMPY:
            return np.max(array)
        elif rc_type == TORCH:
            return torch.max(array)
        elif rc_type == CASADI:
            return casadi.mmax(*safe_unpack(array))

    def sum_2(self, array, rc_type: RCType = NUMPY):
        if isinstance(array, (list, tuple)):
            rc_type = type_inference(*array)

        if rc_type == NUMPY:
            return np.sum(array**2)
        elif rc_type == TORCH:
            return torch.sum(array**2)
        elif rc_type == CASADI:
            return casadi.sum1(array**2)

    def sum(self, array, rc_type: RCType = NUMPY, axis=None):
        if isinstance(array, (list, tuple)):
            rc_type = type_inference(*array)

        if rc_type == NUMPY:
            return np.sum(array, axis=axis)
        elif rc_type == TORCH:
            return torch.sum(array, dim=axis)
        elif rc_type == CASADI:
            if axis is None:
                return casadi.sum2(casadi.sum1(array))
            if axis == 0:
                return casadi.sum1(array)
            if axis == 1:
                return casadi.sum2(array)

    def mean(self, array, rc_type: RCType = NUMPY):
        if isinstance(array, (list, tuple)):
            rc_type = type_inference(*array)

        if rc_type == NUMPY:
            return np.mean(array)
        elif rc_type == TORCH:
            return torch.mean(array)
        elif rc_type == CASADI:
            length = self.max(self.shape(*safe_unpack(array)))
            return casadi.sum1(*safe_unpack(array)) / length

    def force_column(self, argin, rc_type: RCType = NUMPY):
        assert len(argin.shape) <= 2, "Only 1D and 2D arrays are supported."

        if rc_type == CASADI:
            if argin.shape[1] > argin.shape[0] and argin.shape[0] == 1:
                return argin.T
            else:
                return argin
        else:
            return argin.reshape(-1, 1)

    def force_row(self, argin, rc_type: RCType = NUMPY):
        assert len(argin.shape) <= 2, "Only 1D and 2D arrays are supported."

        if rc_type == CASADI:
            if argin.shape[0] > argin.shape[1] and argin.shape[1] == 1:
                return argin.T
            else:
                return argin
        else:
            return argin.reshape(1, -1)

    def cross(self, A, B, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.cross(A, B)
        elif rc_type == TORCH:
            return torch.cross(A, B)
        elif rc_type == CASADI:
            return casadi.cross(A, B)

    def dot(self, A, B, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.dot(A, B)
        elif rc_type == TORCH:
            return torch.dot(A, B)
        elif rc_type == CASADI:
            return casadi.dot(A, B)

    def sqrt(self, x, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.sqrt(x)
        elif rc_type == TORCH:
            return torch.sqrt(x)
        elif rc_type == CASADI:
            return casadi.sqrt(x)

    def shape(self, array, rc_type: RCType = NUMPY):
        if rc_type == CASADI:
            return array.size()
        elif rc_type == NUMPY:
            return np.shape(array)
        elif rc_type == TORCH:
            return array.size()

    def function_to_lambda_with_params(
        self, function_to_lambda, *params, var_prototype=None, rc_type: RCType = NUMPY
    ):
        if rc_type in (NUMPY, TORCH):
            if params:
                return lambda x: function_to_lambda(x, *params)
            else:
                return lambda x: function_to_lambda(x)
        else:
            try:
                x_symb = self.array_symb(self.shape(var_prototype))
            except NotImplementedError:
                x_symb = self.array_symb((*safe_unpack(self.shape(var_prototype)), 1))

            if params:
                return function_to_lambda(x_symb, *safe_unpack(params)), x_symb
            else:
                return function_to_lambda(x_symb), x_symb

    def lambda2symb(self, lambda_function, *x_symb, rc_type: RCType = NUMPY):
        return lambda_function(*x_symb)

    def torch_tensor(self, x, requires_grad=True, rc_type: RCType = NUMPY):
        return torch.tensor(x, requires_grad=requires_grad)

    def add_torch_grad(x, rc_type: RCType = NUMPY):
        if rc_type == TORCH:
            x.requires_grad = True
        else:
            raise TypeError("Cannot assign grad to non-torch type variable")

    def tanh(self, x, rc_type: RCType = NUMPY):
        if rc_type == CASADI:
            res = casadi.tanh(x)
        elif rc_type == NUMPY:
            res = np.tanh(x)
        elif rc_type == TORCH:
            res = torch.tanh(x)
        return res

    def if_else(self, c, x, y, rc_type: RCType = NUMPY):
        if rc_type == CASADI:
            res = casadi.if_else(c, x, y)
            return res
        elif rc_type == TORCH or rc_type == NUMPY:
            return x if c else y

    def kron(self, A, B, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.kron(A, B)
        elif rc_type == TORCH:
            return torch.kron(A, B)
        elif rc_type == CASADI:
            return casadi.kron(A, B)

    def array_symb(
        self, tup=None, literal="x", rc_type: RCType = NUMPY, prototype=None
    ):
        if prototype is not None:
            shape = self.shape(prototype)
        else:
            shape = tup

        if isinstance(shape, tuple):
            if len(tup) > 2:
                raise ValueError(
                    f"Not implemented for number of dimensions greater than 2. Passed: {len(tup)}"
                )
            else:
                return casadi.MX.sym(literal, *tup)

        elif isinstance(tup, int):
            return casadi.MX.sym(literal, tup)

        else:
            raise TypeError(
                f"Passed an invalide argument of type {type(tup)}. Takes either int or tuple data types"
            )

    def norm_1(self, v, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.linalg.norm(v, 1)
        elif rc_type == TORCH:
            return torch.linalg.norm(v, 1)
        elif rc_type == CASADI:
            return casadi.norm_1(v)

    def norm_2(self, v, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.linalg.norm(v, 2)
        elif rc_type == TORCH:
            return torch.linalg.norm(v, 2)
        elif rc_type == CASADI:
            return casadi.norm_2(v)

    def logic_and(self, a, b, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.logical_and(a, b)
        elif rc_type == TORCH:
            return torch.logical_and(a, b)
        elif rc_type == CASADI:
            return casadi.logic_and(a, b)

    def to_np_1D(self, v, rc_type: RCType = NUMPY):
        if rc_type == CASADI:
            return v.T.full().flatten()
        else:
            return v

    def squeeze(self, v, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.squeeze(v)
        elif rc_type == TORCH:
            return torch.squeeze(v)
        elif rc_type == CASADI:
            assert (
                v.shape[0] == 1 or v.shape[1] == 1
            ), "Only columns and rows are supported."
            if v.shape[0] == 1 and v.shape[1] > 1:
                return v.T
            else:
                return v

    def uptria2vec(self, mat, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            result = mat[np.triu_indices(self.shape(mat)[0])]
            return result
        elif rc_type == TORCH:
            result = mat[torch.triu_indices(self.shape(mat)[0])]
            return result
        elif rc_type == CASADI:
            n = self.shape(mat)[0]

            vec = rg.zeros((int(n * (n + 1) / 2)), prototype=mat)

            k = 0
            for i in range(n):
                for j in range(i, n):
                    vec[k] = mat[i, j]
                    k += 1

            return vec

    def append(self, array, to_append, rc_type: RCType = NUMPY):
        if rc_type == NUMPY:
            return np.append(array, to_append)

    # TODO: DO WE REALLY NEED THIS DM? WHY NOT TO USE HUMAN READABLE TERMINOLOGY? SAY CASADI_NUMERIC, CASADI_SYMB, CASADI_MATSYMB
    @staticmethod
    def DM(mat):
        return casadi.DM(mat)

    @staticmethod
    def SX(mat):
        return casadi.SX(mat)

    @staticmethod
    def MX(mat):
        return casadi.MX(mat)

    def autograd(self, function, x, *args, rc_type: RCType = NUMPY):
        return casadi.Function(
            "f", [x, *args], [casadi.gradient(function(x, *args), x)]
        )

    def to_casadi_function(
        self, symbolic_expression, *symbolic_vars, rc_type: RCType = NUMPY
    ):
        return casadi.Function("f", list(symbolic_vars), [symbolic_expression])

    def soft_abs(self, x, a=20, rc_type: RCType = NUMPY):
        return a * rg.abs(x) ** 3 / (1 + a * x**2)


rg = RCTypeHandler()


def set_seed(seed):
    torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)


def calculate_value(runnning_objectives, timestamps, discount_factor, sampling_time):
    return sum(runnning_objectives * discount_factor**timestamps) * sampling_time



"""Contains blueprint of config for all optimizable objects."""

from typing import Optional, Any, Type, Dict
import torch
import casadi


class OptimizerConfig:
    """Base class for config of optimizable objects."""

    def __init__(
        self,
        kind: str,
        opt_method: Optional[Any] = None,
        opt_options: Optional[dict] = None,
        log_options: Optional[dict] = None,
        config_options: Optional[dict] = None,
    ) -> None:
        """Instantiate OptimizerConfig object.

        Args:
            kind (str): The optimization kind. Can be either of
                ["symbolic", "tensor", "numeric"]
            opt_method (Optional[Any], optional): What method to use.
                For `kind="tensor"` this is `torch.optim.Adam`, for
                instance. For kind="symbolic" this is "ipopt", defaults
                to None
            opt_options (Optional[dict], optional): Options to pass to
                the optimizer. For `kind="tensor"` this is `{"lr":
                0.001}`, for instance. For `kind="symbolic"` this is
                `{"print_level": 0}`, defaults to None
            log_options (Optional[dict], optional): Needed only for
                `kind="symbolic"`, defaults to None
            config_options (Optional[dict], optional): Other global
                options. `n_epochs`, `data_buffer_sampling_method`,
                `data_buffer_sampling_kwargs`, etc., defaults to None
        """
        self.kind = kind
        self.opt_method = opt_method
        self.opt_options = opt_options if opt_options is not None else {}
        self.log_options = log_options if log_options is not None else {}
        self.config_options = config_options if config_options is not None else {}


class TorchOptimizerConfig(OptimizerConfig):
    """Config for torch-based optimizers."""

    def __init__(
        self,
        n_epochs: int,
        data_buffer_iter_bathes_kwargs: Dict[str, Any],
        opt_method_kwargs: Dict[str, Any],
        opt_method: Type[torch.optim.Optimizer] = torch.optim.Adam,
        is_reinstantiate_optimizer: bool = True,
        n_epochs_per_constraint: Optional[int] = None,
    ) -> None:
        """Instantiate TorchOptimizerConfig object.

        Args:
            n_epochs (int.): How many epochs to use during optimization.
            data_buffer_iter_bathes_kwargs (Dict[str, Any]): kwargs for
                `DataBuffer.iter_batches`.
            opt_method_kwargs (Dict[str, Any]): What options to pass to
                the optimizer
            opt_method (Type[torch.optim.Optimizer], optional): What
                optimizer method to use, defaults to torch.optim.Adam
            is_reinstantiate_optimizer (bool, optional): Whether to
                reinstantiate optimizer every time `optimize()` is
                called, defaults to True
            n_epochs_per_constraint (Optional[int], optional): How many
                gradient steps to take to find a constraint feasible
                domain, defaults to None
        """
        super().__init__(
            kind="tensor",
            opt_method=opt_method,
            opt_options=opt_method_kwargs,
            log_options=None,
            config_options={
                "n_epochs": n_epochs,
                "data_buffer_sampling_method": "iter_batches",
                "data_buffer_sampling_kwargs": data_buffer_iter_bathes_kwargs,
                "is_reinstantiate_optimizer": is_reinstantiate_optimizer,
            }
            | (
                {
                    "constrained_optimization_policy": {
                        "is_activated": True,
                        "defaults": {
                            "n_epochs_per_constraint": n_epochs_per_constraint
                        },
                    }
                }
                if n_epochs_per_constraint is not None
                else {}
            ),
        )


class CasadiOptimizerConfig(OptimizerConfig):
    """Config for casadi-based optimizers."""

    def __init__(
        self,
        batch_size=1,
    ) -> None:
        """Instantiate CasadiOptimizerConfig object.

        Args:
            batch_size (int, optional): How many latest samples to use
                from `DataBuffer`, defaults to 1
        """
        super().__init__(
            kind="symbolic",
            opt_method="ipopt",
            opt_options={"print_level": 0},
            log_options={"print_in": False, "print_out": False, "print_time": False},
            config_options={
                "data_buffer_sampling_method": "sample_last",
                "data_buffer_sampling_kwargs": {
                    "n_samples": batch_size,
                    "dtype": casadi.DM,
                },
            },
        )


class ScipyOptimizerConfig(OptimizerConfig):
    def __init__(
        self,
        kind: str = "numeric",
        opt_method: Optional[Any] = "SLSQP",
        opt_options: Optional[Dict] = None,
        log_options: Optional[Dict] = None,
        config_options: Optional[Dict] = None,
    ) -> None:
        super().__init__(kind, opt_method, opt_options, config_options)
